#!/usr/bin/python

# ch341_control.py
#       --copyright--                   Copyright 2023 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       April 14, 2023          bar
#       --eodstamps--
##      \file
#       \namespace              tzpython.ch341_control
#
#
#   Original   code  from   https://gist.github.com/RJ/7acba5b06a03c9b521601e08d0327d56
#   Info about proto from   https://www.amazon.com/dp/B075GLQQC9?psc=1&ref=ppx_yo2ov_dt_b_product_details
#
#   I got the ch341 running by  "sudo apt remove brltty" the Brial TTY, I assume. The Internet says "autoremove" rather than "remove".
#       An Internet note says browser functionality like "teams" doesn't work, then.
#   I don't find any /dev/ttyUSB* devices, as some do at 9600 baud.
#       Maybe after a reboot, I'll see a ttyUSB.
#   I added     /etc/udev/rules.d/94-ch341usb-relay.rules   which has:
#       SUBSYSTEM=="usb", ATTR{idVendor}=="1a86",ATTR{idProduct}=="7523", MODE="0666"
#
#   To run with pyusb debugging:
#       PYUSB_DEBUG=debug python relay.py
#
#   Grab the vendor and product codes from syslog when plugging in the relay (use lsusb to list all usb devices):
#
#   usb 3-1: New USB device found, idVendor=1a86, idProduct=7523
#
#                       start byte
#                             relay 1...
#                                   state 1=/---\    0=/-\--
#                                         checksum
#   close_relay_cmd = [ 0xA0, 0x01, 0x01, 0xA2 ]
#   open_relay_cmd  = [ 0xA0, 0x01, 0x00, 0xA1 ]    # initial state, light off
#
#
"""
python  %s (options) input_file(s)...
Version %s

  CH341 usb-controlled relay board program.

  Options:

    --loop_for      seconds     Only loop, toggling the relays, for this many seconds.
                                (default: forever)

    --verbose                   Raise the verbosity level.

    --version                   Print the program version, current date/time
                                  and exit.

    @file_name                  Include command line options given in this file.
                                    Semi-colons start comment lines in the file.
                                Note: The file looks like this:

                                      ; ----------------------------------
                                      ; %s.py options

                                      --verbose

                                      ; eof
    Note: If the file,
    %s
    exists, I add it as an '@' file to the beginning of the command line options.

"""

from    __future__  import  print_function


import  os
import  sys
import  time
import  usb.core
import  usb.util

import  tzlib


VENDOR  = 0x1a86
PRODUCT = 0x7523


__version__ = "0.0.0"


MY_NAME = os.path.splitext(os.path.basename(__file__))[0]
if  not MY_NAME :
    MY_NAME = "TEMPLATE"


def oops(err_msg) :
    sys.stderr.write(err_msg + "\n")
    sys.exit(1)


def get_config_file_name() :
    """ Return the user's default config file name. """
    return(tzlib.expand_user_vars(os.path.join('~', '.' + MY_NAME, MY_NAME + '.cfg')))


def get_args(argv, program_name)  :
    """ Yeah, well, the Python world is so filled with getargs modules. I'd like a pop up CUI to be generated, myself, driven from the available options. And infinite history for each field and for all options as a group, and etc. """
    argv                = list(argv)

    args                = tzlib.an_object('an_args')
    args.verbose        = 0                                                                 # how printy we are
    args.loop_time      = sys.maxsize                                                       # for tz_lib_test
    args.args           = []                                                                # unrecognized arguments
    args.file_names     = []

    cfg_file_name       = os.path.abspath(get_config_file_name())                           # config file to add run-time command line options to

    #
    #   Print HELP if just running the program with no options, etc.
    #       if  (not len(argv)) or (tzlib.array_find([ a.lower() for a in argv ], [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0) :
    #
    if  tzlib.find_argi_any_case(argv, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ]) >= 0 :
        bpn             = os.path.basename(program_name or (MY_NAME + ".py"))
        print(__doc__ % ( bpn, __version__, MY_NAME, cfg_file_name, ))
        sys.exit(0)                         # well, there goes my 254 :)

    import  TZCommandLineAtFile
    if  os.path.isfile(      cfg_file_name) :
        argv.insert(0, "@" + cfg_file_name)
    TZCommandLineAtFile.expand_at_sign_command_line_files(argv, encoding = 'utf8')

    while len(argv)     :
        aa  = argv.pop(0)
        a   = aa.lower()

        try :
            if  tzlib.find_argi(a, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0 :
                oops("Please keep --help /help -h -? /? -? /h out of configuration/option files!")
                sys.exit(1)

            if  tzlib.find_argi(a, [ "--verbose", "-v", ]) >= 0 :
                args.verbose       += 1
                continue

            if  tzlib.find_argi(a, [ "--version", ]) >= 0 :
                print(MY_NAME + " version " + __version__ + " run at " + time.asctime())
                sys.exit(0)

            if  tzlib.find_argi(a, [ "--loop_time", "--loop_for", "--loop_duration", ]) >= 0 :
                args.loop_time      = float(argv.pop(0))
                continue

            if  a and (a[0] != '-') :
                args.file_names.append(aa)
                continue

            oops("I cannot understand %s!" % aa)

        except IndexError :
            oops("I need more information for %s !" % aa)
        except ( TypeError, ValueError, ) :
            oops("I need better information for %s - a number, perhaps!" % aa)
        pass

    return(args)



class   a_board(object) :
    """ Handle a CH340/CH341 USB relay control board. """

    def make_board_started(me) :
        """
            If we don't have a running board, find one and start it if we can.

            This can be called while the board is not plugged in or when something is wrong.
            But, really, if such is the case, you might consider sleeping for a while between calls.

            Raise ValueError if we can't find a USB device.

            Raise IOError if things are not good. Maybe the device is in use by someone else.

            Return True if all is well with the board, so far as we know (we do not interact with it).
        """
        if  not me.brd :
            v       = VENDOR
            p       = PRODUCT

            dev     = usb.core.find(idVendor = v, idProduct = p)

            if  dev is None:
                raise ValueError("Device %04x:%04x not found" % ( v, p, ))

            if  dev.is_kernel_driver_active(0):
                dev.detach_kernel_driver(0)

            cfg     = dev.get_active_configuration()
            intf    = cfg[(0,0)]

            me.brd  = usb.util.find_descriptor(
                        intf,
                        custom_match = lambda e : usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_OUT,     # match the first OUT endpoint
                                              )
            if  not me.brd :
                me.brd  = None
                raise IOError("Device %04x:%04x perhaps already started by someone else?" % (v, p, ))
            pass
        return(True)


    def __init__(me) :
        """ Initialize a CH340/CH341 USB relay board. """
        me.brd  = None
        try     :
            me.make_board_started()
        except ( ValueError, IOError, ) :
            me.brd  = None                      # suspenders for our belt
            pass                                # let him get us going and then find out about the problem when he tries to set a relay or when he notices our 'brd' is None
        pass


    def relay(me, n = 1, lit = False, verbose = 0) :
        if  not (0 < n <= 16) :
            return(IndexError("No such relay %i (should be 1..16)" % n))

        if  not me.brd :
            me.brd  = None
            raise IOError("No device for relay()")

        cmd     = [ 0xa0, n, (lit and 1) or 0, ]            # lit refers to the LED on the board. If it's lit, the middle connecter is disconnected and the outside one is connected.
        cmd.append(sum(cmd))                                # put a checksum on the end of what we send the device
        err     = False
        try     :
            for i in range(16) :                            # 16 is more reliable than 1 - I've not tried any other values
                me.brd.write(cmd)
            if  verbose :
                print("%i %s" % ( n, (lit and "lit") or "off"))
            pass
        except usb.core.USBError :
            me.brd  = None
            err     = True

        if  err     :
            raise IOError("Device with relay %i probably removed" % n)

        pass


    #   a_board



if  __name__ == "__main__":

    program_name    = sys.argv[0]
    args            = get_args(sys.argv[1:], program_name)

    me              = a_board()
    brd             = me.brd
    if  not brd     :
        print("No board found or started")
    dev             = brd

    start_time      = time.time()

    while time.time() - start_time < args.loop_time :
        try         :
            try     :
                me.make_board_started()             # just in case
                if  not brd :
                    brd = me.brd
                    print("Board started")
                pass
            except  ValueError :
                if  brd :
                    brd = me.brd
                    print("Board is not found.")
                pass
            except  IOError :
                if  brd :
                    brd = me.brd
                    print("Board probably in use.")
                pass

            if  not me.brd :
                time.sleep(0.3)                     # let's not hog the CPU
            else    :
                me.relay(1, False, verbose = args.verbose)
                time.sleep(1)
                me.relay(2, True,  verbose = args.verbose)
                time.sleep(1)
                me.relay(1, True,  verbose = args.verbose)
                time.sleep(1)
                me.relay(2, False, verbose = args.verbose)
                time.sleep(2)
                dev = me.brd
            pass
        except (ValueError, IOError, ) :
            if  dev :
                dev = None
                print("Device removed or in use")
            pass
        except KeyboardInterrupt :
            try :
                print('')
                print("By your command, oh great keyboard in the sky, I stop.")
                me.relay(1, False, verbose = args.verbose)
                me.relay(2, False, verbose = args.verbose)
                print('')
            except  IndexError :
                print("Hey, what's wrong with relay 1 or 2!?!")
            except ( ValueError, IOError, ) :
                print("Device removed or in use")
            print('')
            break
        pass
    pass


# eof
