#!/usr/bin/python

# find_gps_tracks_to_combine.py
#       --copyright--                   Copyright 2023 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       June 4, 2023            bar
#       June 6, 2023            bar     format the separete-line milage number for easier sorting
#       --eodstamps--
##      \file
#       \namespace              tzpython.find_gps_tracks_to_combine
#
#       TODO:
#           using the lat/lon limits of the tracks, don't look in tracks which are a minimum distance from the point we're looking for near points to if the best point found so far is less than that min distance
#           for no real reason, index through the points using an index step of 3000017, a prime number bigger than we'll ever see track point arrays
#           print the closest point pair so we can check the shortest distance between two geotracks
#
#
"""
python  %s (options) input_file(s)...
Version %s

  Find the nearest track to the given tracks.
  This helps find tracks to combine into geotracks (combos of nearby tracks).

  Options:

    --search_for    file_name   Search the input_files for closest to this file.
                                  Put ambiguous --search_for file name in quotes.

    --ignore        file_name   Do not search for this file or files.
                                  Put ambiguous --ignore file name in quotes.

    --test                      For testing, use no randomization.

    --verbose                   Raise the verbosity level.

    --version                   Print the program version, current date/time
                                  and exit.

    @file_name                  Include command line options given in this file.
                                    Semi-colons start comment lines in the file.
                                Note: The file looks like this:

                                      ; ----------------------------------
                                      ; %s.py options

                                      --verbose

                                      ; eof
    Note: If the file,
    %s
    exists, I add it as an '@' file to the beginning of the command line options.

"""

from    __future__  import  print_function

import  glob
import  os
import  random
import  sys
import  time

import  numpy

import  latlon
import  tz_gps
import  tzlib


__version__ = "0.0.0"


MY_NAME = os.path.splitext(os.path.basename(__file__))[0]
if  not MY_NAME :
    MY_NAME = "find_gps_tracks_to_combine"


def oops(err_msg) :
    sys.stderr.write(err_msg + "\n")
    sys.exit(1)


def get_config_file_name() :
    """ Return the user's default config file name. """
    return(tzlib.expand_user_vars(os.path.join('~', '.' + MY_NAME, MY_NAME + '.cfg')))


def get_args(argv, program_name)  :
    """ Yeah, well, the Python world is so filled with getargs modules. I'd like a pop up CUI to be generated, myself, driven from the available options. And infinite history for each field and for all options as a group, and etc. """
    argv                = list(argv)

    args                = tzlib.an_object('an_args')
    args.verbose        = 0                                                                 # how printy we are
    args.args           = []                                                                # unrecognized arguments
    args.file_names     = []
    args.search_for     = []                                                                # optional separate list of files to look for closest files to
    args.ignores        = []                                                                # optional separate list of files to ignore
    args.test           = False                                                             # allow randomized operation

    cfg_file_name       = os.path.abspath(get_config_file_name())                           # config file to add run-time command line options to

    #
    #   Print HELP if just running the program with no options, etc.
    #       if  (not len(argv)) or (tzlib.array_find([ a.lower() for a in argv ], [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0) :
    #
    if  tzlib.find_argi_any_case(argv, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ]) >= 0 :
        bpn             = os.path.basename(program_name or (MY_NAME + ".py"))
        print(__doc__ % ( bpn, __version__, MY_NAME, cfg_file_name, ))
        sys.exit(0)                         # well, there goes my 254 :)

    import  TZCommandLineAtFile
    if  os.path.isfile(      cfg_file_name) :
        argv.insert(0, "@" + cfg_file_name)
    TZCommandLineAtFile.expand_at_sign_command_line_files(argv, encoding = 'utf8')

    while len(argv)     :
        aa  = argv.pop(0)
        a   = aa.lower()

        try :
            if  tzlib.find_argi(a, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0 :
                oops("Please keep --help /help -h -? /? -? /h out of configuration/option files!")
                sys.exit(1)

            if  tzlib.find_argi(a, [ "--verbose", "-v", ]) >= 0 :
                args.verbose       += 1
                continue

            if  tzlib.find_argi(a, [ "--version", ]) >= 0 :
                print(MY_NAME + " version " + __version__ + " run at " + time.asctime())
                sys.exit(0)

            if  tzlib.find_argi(a, [ "--search_for", "--search", "-s", ]) >= 0 :
                args.search_for.append(argv.pop(0))
                continue

            if  tzlib.find_argi(a, [ "--ignore", "--ig", ]) >= 0 :
                args.ignores.append(argv.pop(0))
                continue

            if  tzlib.find_argi(a, [ "--test", ]) >= 0 :
                args.test   = True
                continue

            if  a and (a[0] != '-') :
                args.file_names.append(aa)
                continue

            oops("I cannot understand %s!" % aa)

        except IndexError :
            oops("I need more information for %s !" % aa)
        except ( TypeError, ValueError, ) :
            oops("I need better information for %s - a number, perhaps!" % aa)
        pass

    return(args)



def read_track_files(hdr, file_names, verbose = 0) :
    """ Read in the given GPS tracks to a dictionary keyed by file name, valued by an array of a_track()s in the file. """
    indent          = ""
    if  verbose and hdr :
        print(hdr)
        indent      = "  "
    files           = {}
    tcnt            = 0
    pcnt            = 0
    for afn in file_names :
        fns = glob.glob(afn)
        if  not len(fns) :
            oops("No files found for %s!" % afn)
        for fn in fns :
            abfn    = os.path.abspath(fn)
            if  abfn not in files :
                if  verbose > 1 :
                    print("%s%s" % ( indent, fn, ), end = ' ')
                    sys.stdout.flush()
                trks        = tz_gps.tracks_from_file(fn)
                tcnt       += len(trks)
                pcnt       += tz_gps.count_points_in_tracks(trks)
                files[abfn] = trks
                if  verbose > 1 :
                    print("%s  %u tracks" % ( indent, len(trks), ))
                pass
            pass
        pass
    if  verbose :
        print("%s%u tracks %u points" % ( indent, tcnt, pcnt, ))
        print('')
    return(files)


def add_xyz_to_points(tracks) :
    """ Precompute all the X Y Z values for all these tracks' points. """
    for t in tracks :
        for p in t.points :
            p._XYZ  = numpy.array(p.x_y_z(), float)
        pass
    pass


def add_min_max_to_tracks(tracks) :
    """ Precompute the extents of all the tracks (so we can ignore tracks that are too far away) """
    for t in tracks :
        pa          = [ p.lat             for p in t.points ]
        t.min_lat   = min(pa)
        t.max_lat   = max(pa)
        pa          = [ p.lon             for p in t.points ]
        t.min_lon   = min(pa)
        t.max_lon   = max(pa)
        pa          = [ p.altitude or 0.0 for p in t.points ]
        t.min_alt   = min(pa)
        t.max_alt   = max(pa)
    pass


def tracks_maybe_closer_than(t1, t2, max_d) :
    """ Is it possible for two tracks to have closer points than the given distance? """
    max_d      += 1.0                                               # give us some slack (a nautical mile) just to make sure
    lat_over    = (t1.min_lat <= t2.max_lat) and (t1.max_lat >= t2.max_lat)
    lon_over    = (t1.min_lon <= t2.max_lon) and (t1.max_lon >= t2.max_lon)
    if  lat_over and lon_over :
        return(True)                                                # the tracks overlap so they are close enough
    if  lat_over :
        return(min(latlon.calcDistance(t1.min_lat, t1.max_lon, t2.min_lat, t2.min_lon), latlon.calcDistance(t1.min_lat, t1.min_lon, t2.min_lat, t2.max_lon)) <= max_d)
    if  lon_over :
        return(min(latlon.calcDistance(t1.min_lat, t1.min_lon, t2.max_lat, t2.min_lon), latlon.calcDistance(t1.max_lat, t1.min_lon, t2.min_lat, t2.min_lon)) <= max_d)

    for t1lat               in ( t1.min_lat, t1.max_lat ) :         # for the 4 corners of each track (do all combos to handle date line and poles, sorta
        for t1lon           in ( t1.min_lon, t1.max_lon ) :
            for t2lat       in ( t2.min_lat, t2.max_lat ) :
                for t2lon   in ( t2.min_lon, t2.max_lon ) :
                    d   = latlon.calcDistance(t1lat, t1lon, t2lat, t2lon)
                    if  d < max_d :                                 # if the corner pair is closer than the maximum distance the tracks can be apart
                        return(True)
                    pass
                pass
            pass
        pass
    return(False)



if  __name__ == '__main__' :
    program_name    = sys.argv[0]
    args            = get_args(sys.argv[1:], program_name)

    sfiles          = []
    if  args.search_for :
        sfiles      = read_track_files("Search...",  args.search_for, verbose = args.verbose)
    files           = read_track_files("Files...",   args.file_names, verbose = args.verbose)
    ignores         = {}
    if  args.ignores :
        ignores     = read_track_files("Ignores...", args.ignores,    verbose = args.verbose)
        for fn in ignores.keys() :
            files.pop(fn, None)
        pass
    if  not sfiles  :
        sfiles      = files

    for tracks in files.values() :
        add_xyz_to_points(tracks)
    for tracks in sfiles.values() :
        add_xyz_to_points(tracks)
    for tracks in files.values() :
        add_min_max_to_tracks(tracks)
    for tracks in sfiles.values() :
        add_min_max_to_tracks(tracks)

    for fn, tracks in sorted(sfiles.items(), key = lambda a : a[0].upper()) :
        bt  = None
        bd  = float(sys.maxsize)
        bds = float(sys.maxsize)
        if  args.verbose > 1 :
            print('Finding nearest to "%s"' % os.path.basename(fn))
        if  not args.test :
            random.shuffle(tracks)
        for t in tracks :                               # for each track to find a close point to
            fn_fls  = [ ( tfn, trks ) for tfn, trks in files.items() if tfn != fn ]
            if  args.test :
                fn_fls.sort()
            else :
                random.shuffle(fn_fls)                  #   because I get bored
            tpa = t.points
            if  not args.test :
                random.shuffle(tpa)
            for p in tpa :                              #   for each point in the track to search for a close point to
                for tfn, trks in fn_fls :               #     for each file to find the point in
                    for tt in trks  :                   #       for each track to find a close point in
                        if  tracks_maybe_closer_than(t, tt, bds) :
                            for pp in tt.points :       #         for each point to see if it's closest to the searched-for point
                                d   = numpy.linalg.norm(pp._XYZ - p._XYZ)
                                if  bd  > d :
                                    bd  = d
                                    bp  = p
                                    bpp = pp
                                    bds = bp.distance_from(bpp)
                                    ftn =  t.name or os.path.basename( t.file_name)
                                    ttn = tt.name or os.path.basename(tt.file_name)
                                    if  bt != tfn :         # don't overprint - at the cost of printing a sub-optimal distance for this file
                                        bt  = tfn
                                        if  args.verbose > 2 :
                                            print("    better %s %f" % ( os.path.basename(bt), bp.distance_from(bpp) * latlon.milesPerNauticalMile, ))  # note: distance is not the closest for the file
                                        pass
                                    pass
                                pass
                            pass
                        pass
                    pass
                pass
            pass
        if  bt :
            print('%s%s in "%s" is %u feet:%.2f miles from %s in "%s"' % (
                                                                            ((args.verbose > 1) and "  ") or "",
                                                                            bp,
                                                                            os.path.basename(fn),
                                                                            bp.distance_from(bpp) * latlon.milesPerNauticalMile * 5280,
                                                                            bp.distance_from(bpp) * latlon.milesPerNauticalMile,
                                                                            bpp,
                                                                            os.path.basename(bt),
                                                                         )
                 )
            if  args.verbose :          # and ((os.path.basename(fn) != ftn) or (os.path.basename(bt) != ttn)) :
                print('%s  %7.2f miles from tracks named: "%s"  and  "%s"' % ( ((args.verbose > 1) and "  ") or "", bp.distance_from(bpp) * latlon.milesPerNauticalMile, ftn, ttn, ))
            if  args.verbose > 2 :
                print('')
            pass
        pass

    pass

#
#
# eof
