#!/usr/bin/python

# gps_dwell.py
#       --copyright--                   Copyright 2016 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       December 8, 2016        bar
#       December 9, 2016        bar     start speeding it up
#       December 11, 2016       bar     remove debug printouts
#       March 5, 2023           bar     future print
#       --eodstamps--
##      \file
#       \namespace              tzpython.gps_dwell
#
#
#       Extract waypoints for times when GPS tracks dwell at some point.
#
#

from    __future__  import  print_function

import  os
import  sys


import  latlon
import  tz_gps
import  tz_parse_time
import  tzlib


OTHER_POINT_MULT    = 14.0


def _find_next_point_inside_radius(pts, i, bp, radius, max_out_points) :
    ii      = i
    while   (ii < len(pts)) and (ii - i <= max_out_points) :
        if  bp.flat_distance_from(pts[ii]) <= radius :
            return(ii)
        ii += 1
    return(0)



def _merge_big_points(pts, radius, by_lat) :
    pts     = list(pts)
    for pi, p in enumerate(pts) :
        pii = pi + 1
        while pii < len(pts) :
            pp  = pts[pii]
            if  p.flat_distance_from_lat_lon(pp.lat if by_lat else p.lat, pp.lon if not by_lat else p.lon) > radius * OTHER_POINT_MULT :
                break
            if  p.flat_distance_from(pp) <= radius * OTHER_POINT_MULT :
                flip        = (pp.when < p.when)
                if  flip    :
                    pts[pi], pts[pii]   = pts[pii], pts[pi]
                    p,       pp         = pp,       p
                pp.when_tot = pp.cnt * p.when
                pp.when     = pp.when_tot / pp.cnt
                p.include_point(pp)
                del(pts[pii])
                if  flip    :
                    pii     = pi + 1                # start over because we've changed the point's time (and location, too, for that matter)
                pass
            else            :
                pii        += 1
            pass
        pass
    return(pts)

def merge_big_points(pts, radius) :
    pts = list(pts)
    pts.sort(lambda a, b : cmp(a.lat, b.lat))
    pts = _merge_big_points(pts, radius, True)
    pts.sort(lambda a, b : cmp(a.lon, b.lon))
    pts = _merge_big_points(pts, radius, False)
    return(pts)


def tracks_to_waypoints(tracks, radius, min_duration, min_points, max_out_points) :
    tracks          = tz_gps.make_array_of_tracks(tracks)
    big_points      = []
    for t in tracks :
        if  len(t.points) :
            bpt     = tz_gps.a_track([], len(big_points), t.points[0].when, name = t.name, file_name = t.file_name, color = t.color)
            bp      = tz_gps.a_big_point(t.points[0])
            i       = 0
            while i < len(t.points) - 1 :
                i  += 1
                p   = t.points[i]
                if  bp.flat_distance_from(p) <= radius :
                    bp.include_point(p)
                else :
                    ii      = _find_next_point_inside_radius(t.points, i + 1, bp, radius, max_out_points)
                    if  ii  :
                        i   = ii - 1                # skip forward to include this found point
                    else    :
                        if  (bp.cnt >= min_points) and (bp.end_when - bp.start_when >= min_duration) :
                            bp._file_name   = bpt.file_name
                            bpt.points.append(bp)
                        bp  = tz_gps.a_big_point(p)
                    pass
                pass
            if  (bp.cnt >= min_points) and (bp.end_when - bp.start_when >= min_duration) :
                bp._file_name   = bpt.file_name
                bpt.points.append(bp)
            if  len(bpt.points) :
                big_points.append(bpt)
            pass
        pass
    return(tz_gps.points_from_all_tracks(tz_gps.make_array_of_tracks(big_points)))


def optimize_waypoints(pts) :
    pts = list(pts)
    for p in pts    :
        p.points    = []            # stop us from taking too much memory
        p.id_points = {}
    # pts.sort(lambda a, b : cmp(a.when, b.when) or cmp(os.path.basename(a._file_name).lower(), os.path.basename(b._file_name).lower()) or cmp(a._file_name.lower(), b._file_name.lower()) or cmp(os.path.abspath(a._file_name).lower(), os.path.abspath(b._file_name).lower()))
    pts = merge_big_points(pts, radius)
    pts.sort(lambda a, b : cmp(os.path.basename(a._file_name).lower(), os.path.basename(b._file_name).lower()) or cmp(a._file_name.lower(), b._file_name.lower()) or cmp(os.path.abspath(a._file_name).lower(), os.path.abspath(b._file_name).lower()) or cmp(a.when, b.when))
    return(pts)



help_str    = """
%s  (options) gpx_kmz_kml_file(s)

Extract waypoints from when GPS tracks dwell at some location for a time.

Options:

    --output            file_name   Output the waypoints to the given .gpx/.kml/.kmz file.
    --radius            meters      Dwell points must all be inside the given radius from their center.    (default: %u)
    --min_time          hh:mm:ss    Dwell points must include points encompassing at least this duration.  (default: %s)
    --min_points        count       Dwell points must include this number of GPS sample points.            (default: %u)
    --max_out_points    count       Allow only this number of points outside the bounds of space and time. (default: %u)

    --browse_file       file_name   Put links to Google Maps for each waypoint in a .htm file and open the file in a browser.
    --verbose                       Increase the verbosity level.


"""

#
#
#   Main program
#
#
if  __name__ == '__main__' :

    import  gc
    import  subprocess

    import  TZCommandLineAtFile
    import  tz_browser
    import  tz_google_earth
    import  tz_os_priority


    program_name    = sys.argv.pop(0)
    TZCommandLineAtFile.expand_at_sign_command_line_files(sys.argv)


    radius          = 50.0 / latlon.metersPerNauticalMile
    min_time        = 5 * 60.0
    min_points      = int(min_time / 60)
    max_out_points  = 5
    ofile_name      = None
    browse_file     = None
    verbose         = 0


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--help", "-h", "-?", "/h", "/?", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        print(help_str % ( program_name, radius * latlon.metersPerNauticalMile, tzlib.wdhms_str(min_time), min_points, max_out_points, ), file = sys.stderr)
        sys.exit(254)


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--verbose", "-v", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        verbose    += 1


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--browser_file", "--browser-file", "--browserfile", "--bf", "-b", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        browse_file   = tzlib.expand_user_vars(sys.argv.pop(oi))


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--radius", "-r", ] )
        if  oi < 0 :    break
        del sys.argv[oi]

        radius  = float(sys.argv.pop(oi)) / latlon.metersPerNauticalMile


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--min_time", "--min-time", "--mintime", "--min_duration", "--min-duration", "--minduration", "--md", "-t", ] )
        if  oi < 0 :    break
        del sys.argv[oi]

        min_time     = tz_parse_time.parse_time_zone(sys.argv.pop(oi))
        if  min_time == None :
            print("I cannot understand the minimum duration!")
            sys.exit(102)
        pass


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--min_points", "--min-points", "--minpoints", "--mp", ] )
        if  oi < 0 :    break
        del sys.argv[oi]

        min_points = int(sys.argv.pop(oi))


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--max_out_points", "--max-out-points", "--maxoutpoints", "--mop", ] )
        if  oi < 0 :    break
        del sys.argv[oi]

        max_out_points  = int(sys.argv.pop(oi))


    while True :
        oi  = tzlib.find_argi(sys.argv, [ "--out_put", "--out-put", "--output", "--out_file", "--out-file", "--outfile", "--output_file", "--output-file", "--outputfile", "-o", ] )
        if  oi < 0 :    break
        del sys.argv[oi]

        ofile_name      = sys.argv.pop(oi)


    tz_os_priority.set_proc_to_idle_priority()

    stime   = tzlib.elapsed_time()

    tcnt    = 0
    wpts    = []
    dcnt    = 0
    try     :
        for fn in sys.argv :
            fns = tzlib.ambiguous_file_list(fn, do_sub_dirs = False)
            for fnn in fns :
                if  verbose > 1 :
                    print(fnn)
                ta  = tz_gps.tracks_from_file(fnn)                  # note: the file name can be converted to a file name in metadata in a .gpx file - an abspath file name, in practice at least one time: muriel_fresh_and_easy_01x.gpx
                if  not ta  :
                    print("%sNo tracks in %s!" % ( ((verbose > 1) and "  ") or "", fnn, ))
                else        :
                    tcnt   += len(ta)
                    wpts   += optimize_waypoints(tracks_to_waypoints(ta, radius, min_time, min_points, max_out_points))
                    ta      = None
                    dcnt   += 1
                    if  dcnt   >= 43 :
                        dcnt    = 0
                        wpts    = optimize_waypoints(wpts)
                        gc.collect()
                    pass
                pass
            pass
        pass
    except KeyboardInterrupt :
        pass
    wpts    = optimize_waypoints(wpts)

    wpts.sort(lambda a, b : cmp(a.when, b.when) or cmp(os.path.basename(a._file_name).lower(), os.path.basename(b._file_name).lower()) or cmp(a._file_name.lower(), b._file_name.lower()) or cmp(os.path.abspath(a._file_name).lower(), os.path.abspath(b._file_name).lower()))

    if  verbose :
        print("%u track%s, %u dwelling waypoint%s." % ( tcnt, tzlib.s_except_1(tcnt), len(wpts), tzlib.s_except_1(len(wpts)), ))

    if  browse_file :
        s   = "<html>\n<body>\n"
        pfn = None
        ps  = ""
        for p in wpts :
            url = tz_google_earth.google_maps_lat_lon_url(p.lat, p.lon)
            if  verbose > 2 :
                print(os.path.basename(p._file_name), url)
            if  pfn    != p._file_name :
                pfn     = p._file_name
                s  += "%s<h2>%s</h2>\n" % ( ps, pfn, )
                ps  = "&nbsp;<p>"
            s  += "&nbsp;&nbsp;&nbsp;&nbsp;%s <a href='%s' target='_blank'>Google Maps</a><br>\n" % ( str(p), url, )
        s      += "\n</body>\n</html>\n"
        tzlib.write_whole_text_file(browse_file, s)
        if  True :
            tz_browser.start_file(browse_file)
        else :
            subprocess.Popen( [ "/usr/bin/chromium-browser", browse_file ] )
        pass

    if  ofile_name :
        tz_gps.write_gps_file(ofile_name, program_name = os.path.basename(__file__), waypoints = wpts, gpx_trk = True)
    else :
        pfn = None
        for p in wpts :
            if  pfn    != p._file_name :
                pfn     = p._file_name
                print(pfn)
            print(" ", p)
        pass

    if  verbose :
        print("Run time:", tzlib.wdhms_str(tzlib.elapsed_time() - stime))

    pass


#
#
# eof
