#!/usr/bin/python

# gps_file_cache.py
#       --copyright--                   Copyright 2022 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       July 18, 2022           bar
#       July 25, 2022           bar     default to caching all gps files in the current directory or in any given directory
#                                       do .jpg files, too
#       July 26, 2022           bar     filter out points and tracks that are at unusual lat/lon 1/1/1970-ish
#                                       v1
#       July 27, 2022           bar     jpg tracks-in-array bug fixed
#                                       improve --version printout
#       July 30, 2022           bar     a_cache() defaults the cache file name
#       February 25, 2023       bar     python3
#       March 1, 2023           bar     use tzlib's BEST_PICKLE_PROTOCOL
#       May 30, 2023            bar     a_cache.forget() though not needed and also, the try/except should be "except KeyboardInterrupt" maybe
#                                       a_cache.print()
#       June 5, 2023            bar     a_file.can_point_be_inside()
#       June 14, 2023           bar     return points array from a_file.add_info()
#       --eodstamps--
##      \file
#       \namespace              tzpython.gps_file_cache
#
#
"""
python  %s (options) input_file(s)...
Version %s

  Keep a cache of metainformation for GPS track files.

  This information includes begin/end times and points, maximum
  point extents in lat/lon/altitude, number of points, etc.

  Store the information under base file name, given file name,
  absolute file name, and a hash of file content.

  input_file(s) defaults to all
  %s files in the current directory unless the --show option is given,
  in which case the cache file is printed whether input files are given
  or not.

  Options:

    --cache_file    file_name   Use the given cache file (a Python pickle file)
                                  default: %s

    --show                      Print the cache file.

    --verbose                   Raise the verbosity level.

    --version                   Print the program version, current date/time
                                  and exit.

    @file_name                  Include command line options given in this file.
                                    Semi-colons start comment lines in the file.
                                Note: The file looks like this:

                                      ; ----------------------------------
                                      ; %s.py options

                                      --verbose

                                      ; eof
    Note: If the file,
    %s
    exists, I add it as an '@' file to the beginning of the command line options.

"""

from    __future__  import  print_function

import  glob
import  os
import  re
import  sys
import  time

import  latlon
import  tz_gps
import  tz_jpg
import  tzlib


__version__ = "0.0.0"


CACHE_VERSION   = 1
CACHE_FILE_EXT  = ".gps_cache"          # base file name of where we store the cache, by default


GPS_FILE_EXTS   = [ ".gpx", ".kmz", ".kml", ".nmea", ".loc", ".jpg", ".jpeg", ]


MY_NAME = os.path.splitext(os.path.basename(__file__))[0]
if  not MY_NAME :
    MY_NAME = "gps_file_cache"


def oops(err_msg) :
    sys.stderr.write(err_msg + "\n")
    sys.exit(1)


def get_config_file_name() :
    """ Return the user's default config file name. """
    return(tzlib.expand_user_vars(os.path.join('~', '.' + MY_NAME, MY_NAME + '.cfg')))

def default_cache_file_name() :
    """ Return the user's default cache file name. """
    return(os.path.splitext(get_config_file_name())[0] + CACHE_FILE_EXT)


def get_args(argv, program_name)  :
    """ Yeah, well, the Python world is so filled with getargs modules. I'd like a pop up CUI to be generated, myself, driven from the available options. And infinite history for each field and for all options as a group, and etc. """
    argv                = list(argv)

    args                = tzlib.an_object('an_args')
    args.verbose        = 0                                                                 # how printy we are
    args.args           = []                                                                # unrecognized arguments
    args.file_names     = []
    args.cache_file     = default_cache_file_name()                                         # where to pickle/store the cache
    args.no_write       = False                                                             # a way for tz_lib_test.py to run this script without it changing the cache
    args.show           = False                                                             # don't print the cache file

    cfg_file_name       = os.path.abspath(get_config_file_name())                           # config file to add run-time command line options to

    #
    #   Print HELP if just running the program with no options, etc.
    #       if  (not len(argv)) or (tzlib.array_find([ a.lower() for a in argv ], [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0) :
    #
    if  tzlib.find_argi_any_case(argv, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ]) >= 0 :
        bpn             = os.path.basename(program_name or (MY_NAME + ".py"))
        print(__doc__ % ( bpn, __version__, re.sub(r"[\[\]\,\']", "", str(GPS_FILE_EXTS)), args.cache_file, MY_NAME, cfg_file_name, ))
        sys.exit(0)                         # well, there goes my 254 :)

    import  TZCommandLineAtFile
    if  os.path.isfile(      cfg_file_name) :
        argv.insert(0, "@" + cfg_file_name)
    TZCommandLineAtFile.expand_at_sign_command_line_files(argv, encoding = 'utf8')

    while len(argv)     :
        aa  = argv.pop(0)
        a   = aa.lower()

        try :
            if  tzlib.find_argi(a, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0 :
                oops("Please keep --help /help -h -? /? -? /h out of configuration/option files!")
                sys.exit(1)

            if  tzlib.find_argi(a, [ "--verbose", "-v", ]) >= 0 :
                args.verbose       += 1
                continue

            if  tzlib.find_argi(a, [ "--version", ]) >= 0 :
                print("%s  version %s  Cache version %s  run at %s" % ( MY_NAME, __version__, CACHE_VERSION, time.asctime(), ))
                sys.exit(0)

            if  tzlib.find_argi(a, [ "--cache_file", "--cache", "--cf", "-c", ]) >= 0 :
                args.cache_file     = argv.pop(0)
                continue

            if  tzlib.find_argi(a, [ "--no_write", "--nw", ]) >= 0 :
                args.no_write       = True
                continue

            if  tzlib.find_argi(a, [ "--show", "--dump", "--show_cache", "--show_cache_file", "--dump_cache", "--dump_cache_file", ]) >= 0 :
                args.show           = True
                continue

            if  a and (a[0] != '-') :
                args.file_names.append(aa)
                continue

            oops("I cannot understand %s!" % aa)

        except IndexError :
            oops("I need more information for %s !" % aa)
        except ( TypeError, ValueError, ) :
            oops("I need better information for %s - a number, perhaps!" % aa)
        pass
    exts    = GPS_FILE_EXTS
    if  (not len(args.file_names)) and (not args.show) :
        for ext in exts :
            args.file_names.append("*" + ext)
        pass
    fns     = {}
    for fn in args.file_names :
        fn  = tzlib.expand_user_vars(fn)
        if  os.path.isdir(fn) :
            for ext in exts :
                for fn in glob.glob(os.path.join(fn, "*" + ext)) :
                    fns[fn] = True
                pass
            pass
        else :
            for fn in glob.glob(fn) :
                fns[fn] = True
            pass
        pass
    fns = list(fns.keys())
    fns.sort()
    args.file_names = fns

    return(args)



class   a_file(object) :
    """ A cache of information about a particular GPS track file. """

    def __init__(me, file_name) :
        """ Create the object without the GPS info and hash. """
        me.file_name    = file_name
        me.abs_name     = os.path.abspath( me.file_name)
        me.base_name    = os.path.basename(me.file_name)
        me.file_when    = os.path.getmtime(me.file_name)
        me.file_size    = os.path.getsize( me.file_name)


    def add_info(me)    :
        """ Add the GPS info. As a favor to outside callers, return an array of all the points in our file. """
        pts = []
        if  not hasattr(me, 'distance') :
            trks                = []
            ext                 = os.path.splitext(me.file_name)[1].lower()
            if  ext not in [ '.jpg', '.jpeg', ] :           # well, tz_gps finds .txt gps tracks in images or about anywhere it looks, so we'll protect jpeg's, but we really should just look only at the 5 gps types
                trks            = tz_gps.tracks_from_file(me.file_name)
                # tz_gps.get_waypoints_from_tracks(trks)      # rids trks of one-point tracks and tracks where all points are waypoints (redundant except it strips points from .km? files gotten from a <LineString>)
                trks            = tz_gps.time_strip_tracks(trks, 10 * 24 * 60 * 60, sys.maxsize)        # remove points before unix time 10 hours.  note: the time start/end period tells which time-points to keep, not strip
                trks            = [ t for t in trks if t and len(t.points) ]
                if  not len(tz_gps.points_from_all_tracks(trks)) :
                    trks        = []
            if  not len(trks)   :
                try             :
                    jpg         = tz_jpg.a_jpg(me.file_name)
                    exif        = jpg.get_exif()
                    if  exif    :
                        lat     = exif.get_latitude()
                        lon     = exif.get_longitude()
                        alt     = exif.get_altitude()
                        t       = exif.picture_taken_time(dflt = os.path.getmtime(me.file_name))
                        if  (lat is not None) and (lon is not None) :
                            trks    = [ tz_gps.a_track([ tz_gps.a_point(when = t, lat = lat, lon = lon, altitude = alt) ]) ]
                        pass
                except ValueError   :
                    trks            = []
                pass

            pts                 = tz_gps.points_from_all_tracks(trks)
            if  not len(pts)    :
                trks            = []                        # don't give 'em anything, really

            me.trk_count        = len(trks)
            me.pts_count        = len(pts)
            pa                  = [ p.lat             for p in pts ]
            me.min_lat          = min(pa)
            me.max_lat          = max(pa)
            pa                  = [ p.lon             for p in pts ]
            me.min_lon          = min(pa)
            me.max_lon          = max(pa)
            pa                  = [ p.altitude or 0.0 for p in pts ]
            me.min_alt          = min(pa)
            me.max_alt          = max(pa)
            pa                  = [ p.when     or 0.0 for p in pts ]
            me.start_time       = min(pa)
            me.end_time         = max(pa)
            me.distance         = tz_gps.tracks_distance(trks)
        return(pts)


    def add_hash(me)        :
        """ Add the hash. """
        if  not hasattr(me, 'hash') :
            d               = tzlib.safe_read_whole_binary_file(me.abs_name)
            me.hash         = tzlib.hashlib.md5(d).hexdigest().lower()
        pass


    def can_point_be_inside(me, plat, plon, slack = None) :
        """ Can the given lat/lon be inside our track bounds? Allow the point to be 'slack' outside in nautical miles if positive and in degrees if negative. """
        slat = slon     = 0.0
        if  slack is not None :
            if  slack   > 0 :                           # is the slack in nautical miles?
                for ll in [
                            ( me.min_lat, me.min_lon, ),
                            ( me.min_lat, me.max_lon, ),
                            ( me.max_lat, me.min_lon, ),
                            ( me.max_lat, me.max_lon, ),
                         ]      :
                    lat, lon    = latlon.distance_in_lat_lon_here(ll[0], ll[1], slack)          # this logic will fail on tracks going over the poles, for instance. but tough luck.
                    slat        = max(slat, lat)
                    slon        = max(slon, lon)
                pass
            elif slack  <  0    :                       # or is it in lat/lon degrees?
                slat    = -slack
                slon    = -slack
            pass
        return((me.min_lat - slat < plat < me.max_lat + slat) and (me.min_lon - slon < plon < me.max_lon + slon))




    PRINT_HEADER = "Trks   Pts MinLatitudeMinLongitude   MaxLatitudeMaxLongitude MinAlt   MaxAlt  Start Time                  End Time                    Miles  Base file name"

    def __str__(me) :
        """ Make a printable string for us. """
        if  not hasattr(me, 'distance') :
            return("                                                                                                                    %s" % me.base_name)
        return("%3u %6u  %10.6f,%-11.6f -> %10.6f,%-11.6f  %5.0f -> %5.0f  %22s -> %-22s  %7.1f  %s" % (
                me.trk_count,
                me.pts_count,
                me.min_lat,
                me.min_lon,
                me.max_lat,
                me.max_lon,
                me.min_alt * latlon.feetPerMeter,
                me.max_alt * latlon.feetPerMeter,
                time.asctime(time.localtime(me.start_time)),
                time.asctime(time.localtime(me.end_time)),
                me.distance * latlon.milesPerNauticalMile,
                me.base_name,
               ))


    #   a_file



class   a_cache(object) :
    """ A cache of information about GPS track files. """

    def __init__(me, file_name = None) :
        me.file_name    = file_name or default_cache_file_name()

        me.dirty        = True          # write it when finished
        me.version      = CACHE_VERSION
        me.d            = {}            # empty cache. keys = file_name and abspath() and basename() and hash    value = a_file()


    def how_many(me) :
        """ Return the number of unique cache entries we have. """
        hd  = {}
        for c in me.d.values() :
            hd[c.hash]  = True
        return(len(hd))


    def forget(me, fn) :
        """ Get rid of any reference to this file. """
        c   = a_file(fn)
        c.add_hash()
        try :                                           # trick to make sure we write out valid data to the cache file when he hits Control C
            me.d.pop(c.file_name, None)
            me.d.pop(c.abs_name,  None)
            me.d.pop(c.base_name, None)
            me.d.pop(c.hash,      None)
            me.dirty    = True
        except :
            me.d.pop(c.file_name, None)
            me.d.pop(c.abs_name,  None)
            me.d.pop(c.base_name, None)
            me.d.pop(c.hash,      None)
            me.dirty    = True
            raise
        pass


    def add(me, fn) :
        """ Insure the given file is in the cache. """
        c   = a_file(fn)
        c.add_hash()                                    # read the file and add the hash to the info for lookup purposes
        ca  = me.get_maybe_in_cache(c)                  # find any matching hashes in the cache
        if  (not len(ca)) or (c.file_name != ca[0].file_name) or (c.abs_name != ca[0].abs_name) or (c.base_name != ca[0].base_name) :
            if  len(ca)         :
                ca[0].file_name = c.file_name           # note, this will leave dangling name references to the updated a_file(). But that's OK. Why not?
                ca[0].abs_name  = c.abs_name
                ca[0].base_name = c.base_name
                c               = ca[0]                 # use the existing a_file()
            else                :
                c.add_info()                            # store the new GPS information (slow)
            try                 :                       # trick to make sure we write out valid data to the cache file when he hits Control C
                me.d[c.file_name]   = c
                me.d[c.abs_name ]   = c
                me.d[c.base_name]   = c
                me.d[c.hash     ]   = c
                me.dirty            = True
            except                  :
                me.d[c.file_name]   = c
                me.d[c.abs_name ]   = c
                me.d[c.base_name]   = c
                me.d[c.hash     ]   = c
                me.dirty            = True
                raise

            return(True)
        return(False)


    def get_maybe_in_cache(me, f) :
        """ Return an array of a_file()'s that this a_file() (with or without a hash computed yet) may be. """
        if  hasattr(f, 'hash') :                        # if he knows the hash to look for,
            c   = me.d.get(f.hash, None)                #    just see if it's in the cache
            return((c and [ c ]) or [])                 # we found the content-matching a_file() or there is none to be found for this GPS file

        fa      = []
        hd      = {}
        for n in [ f.abs_name, f.base_name, f.file_name, ] :
            c   = me.d.get(n, None)
            if  c and (not c.hash in hd) and (f.file_size == c.file_size) :
                hd[c.hash]  = True
                fa.append(c)
            pass
        return(fa)


    def get(me, fn) :
        """ Return a cached a_file() matching this file name, or return None. """
        f   = a_file(fn)
        ca  = me.get_maybe_in_cache(f)
        if  len(ca) :
            return(ca[0])
        return(None)


    def print_to_open_file(me, open_file_to = None) :
        """ Print a list of all we have cached to the given open file or to stdout. """
        fo      = open_file_to or sys.stdout
        hashes  = tzlib.without_dupes([ f.hash  for f in me.d.values() ])
        fls     = [ me.d[h] for h in hashes ]
        fls.sort(key = lambda f : f.abs_name.upper())
        fo.write(a_file.PRINT_HEADER + "\n")
        for f  in fls :
            fo.write(str(f)          + "\n")
        fo.write(a_file.PRINT_HEADER + "\n")


    @staticmethod
    def open(file_name = None) :
        """ Open or create a cache for GPS files. """
        file_name       = file_name or default_cache_file_name()
        me              = tzlib.unpickle_file(file_name)
        if  (not me) or (me.version != CACHE_VERSION) :
            me          = a_cache(file_name)
        else            :
            me.dirty    = False         # we read a valid cache. don't write it unless it's changed (redundant but careful because we write False in .dirty)
        return(me)

    def save(me) :
        """ Write the cache out to its (pickle) file. """
        if  me.dirty    :
            me.dirty    = False
            tzlib.safe_makedirs(os.path.dirname(me.file_name))
            if  not tzlib.pickle_file(me.file_name, me, proto = tzlib.BEST_PICKLE_PROTOCOL) :
                return(False)
            pass
        return(True)
    close   = save                  # keep the old name around


    #   a_cache



if  __name__ == '__main__' :
    program_name    = sys.argv[0]
    args            = get_args(sys.argv[1:], program_name)


    cache           = a_cache.open(args.cache_file)
    def _print_cvln(fs = "") :
        print("%s version %s with %u entries" % ( fs or "Cache", cache.version, cache.how_many(), ))
    _print_cvln()
    hdr             = "Trks  Pnts         South,West          ->        North,East         LowFeet ->  High  Start local when         -> End local when              Miles  File name"
    if  args.verbose :
        print(hdr)

    try :
        for fn in args.file_names :
            if  (cache.add(fn) and args.verbose) or (args.verbose > 1) :
                c   = cache.get(fn)
                if  c.trk_count or c.pts_count   or (args.verbose > 2) :
                    print(cache.get(fn))
                pass
            pass
        pass
    except KeyboardInterrupt :
        args.show = False
        print('')

    if  args.verbose :
        print(hdr)

    if  cache.dirty :
        _print_cvln("Writing cache")

    if  (not args.no_write) and (not cache.save()) :
        oops("Cannnot write %s !" % cache.file_name)

    if  args.show :
        _print_cvln()
        cache.print_to_open_file()
        _print_cvln()
    pass


#
#
# eof
