#!/usr/bin/python

# log_grep.py
#       --copyright--                   Copyright 2021 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       July 14, 2021           bar
#       July 15, 2021           bar     --not
#       February 27, 2023       bar     future print_function
#       March 5, 2023           bar     put the shebang back to python
#       September 27, 2023      bar     --clear_cache
#       September 28, 2023      bar     allow ambiguous log file names with special case for .gz files
#                                       v0.0.2
#       --eodstamps--
##      \file
#       \namespace              tzpython.log_grep
#
#       Handle midnight, when the log files click over to .1 files.
#           If no startswith(), then look for the longest occurance of the current log (equal lengths? go with the 1st one) in the .known log and whack the history before that. Then do the startswith() again.
#
#       NOTE: /etc/hosts.deny is where things are handled on spring
#             iptables seem to be doing things on asuks
#
#
"""
python  %s (options) (regex) ambiguously_named_log_file(s)
Version %s

  Search for a Python regular expression in the part of a Linux log file
  that is new since last we ran.

  Print the lines of text that match the regular expression.

  This is meant to run as a cron job. If it prints nothing, fine.
  If it prints something, the system will email the printout to us.

  Note: if a log file name is not explitely a .gz, I ignore .gz log files.

  Options:

    --find      regex           Print any log file text lines matching this
                                  Python regular expression.
                                  Default: 1st non-option on cmd line.
                                           To ignore the 1st non-option,
                                           set 'regex' to a quoted, empty string.

    --not       regex           Filter out any new lines found that match this
                                  Python regular expression.

    --clear_cache               Don't ignore cached/already-seen finds.

    --verbose                   Raise the verbosity level.

    --version                   Print the program version, current date/time
                                  and exit.

    @file_name                  Include command line options given in this file.
                                    Semi-colons start comment lines in the file.
                                Note: The file looks like this:

                                      ; ----------------------------------
                                      ; %s.py options

                                      --verbose

                                      ; eof
    Note: If the file,
    %s
    exists, I add it as an '@' file to the beginning of the command line options.

"""


from    __future__  import  print_function

import  glob
import  os
import  re
import  sys
import  time

import  tzlib


__version__ = "0.0.2"


MY_NAME = os.path.splitext(os.path.basename(__file__))[0]
if  not MY_NAME :
    MY_NAME = "log_grep"


def oops(err_msg) :
    sys.stderr.write(err_msg + "\n")
    sys.exit(1)


def get_config_file_name() :
    """ Return the user's default config file name. """
    return(tzlib.expand_user_vars(os.path.join('~', '.' + MY_NAME, MY_NAME + '.cfg')))


def get_args(argv, program_name)  :
    """ Yeah, well, the Python world is so filled with getargs modules. I'd like a pop up CUI to be generated, myself, driven from the available options. And infinite history for each field and for all options as a group, and etc. """
    argv                = list(argv)

    args                = tzlib.an_object('an_args')
    args.verbose        = 0                                                                 # how printy we are
    args.args           = []                                                                # unrecognized arguments
    args.file_names     = []
    args.regex          = None                                                              # print any log file text lines that match this regex
    args.not_regex      = None                                                              # except don't print any  lines that match this regex
    args.clear_cache    = False                                                             # if true, don't use cache from previous runs

    cfg_file_name       = os.path.abspath(get_config_file_name())                           # config file to add run-time command line options to

    #
    #   Print HELP if just running the program with no options, etc.
    #       if  (not len(argv)) or (tzlib.array_find([ a.lower() for a in argv ], [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0) :
    #
    if  tzlib.find_argi_any_case(argv, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ]) >= 0 :
        bpn             = os.path.basename(program_name or (MY_NAME + ".py"))
        print(__doc__ % ( bpn, __version__, MY_NAME, cfg_file_name, ))
        sys.exit(0)                         # well, there goes my 254 :)

    import  TZCommandLineAtFile
    if  os.path.isfile(      cfg_file_name) :
        argv.insert(0, "@" + cfg_file_name)
    TZCommandLineAtFile.expand_at_sign_command_line_files(argv, encoding = 'utf8')

    while len(argv)     :
        aa  = argv.pop(0)
        a   = aa.lower()

        try :
            if  tzlib.find_argi(a, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0 :
                oops("Please keep --help /help -h -? /? -? /h out of configuration/option files!")

            if  tzlib.find_argi(a, [ "--verbose", "-v", ]) >= 0 :
                args.verbose       += 1
                continue

            if  tzlib.find_argi(a, [ "--version", ]) >= 0 :
                print(MY_NAME + " version " + __version__ + " run at " + time.asctime())
                sys.exit(0)

            if  tzlib.find_argi(a, [ "--find", "-f", ]) >= 0 :
                if  args.regex is not None :
                    oops("I cannot do two --find regular expressions!")
                args.regex          = argv.pop(0)
                continue

            if  tzlib.find_argi(a, [ "--not", "-n", ]) >= 0 :
                if  args.not_regex is not None :
                    oops("I cannot do two --not regular expressions!")
                args.not_regex      = argv.pop(0)
                continue

            if  tzlib.find_argi(a, [ "--clear_cache", "-c", ]) >= 0 :
                args.clear_cache    = True
                continue

            if  args.regex is None  :
                args.regex          = aa
                continue

            if  a and (a[0] != '-') :
                args.file_names.append(aa)
                continue

            oops("I cannot understand %s!" % aa)

        except IndexError :
            oops("I need more information for %s !" % aa)
        except ( TypeError, ValueError, ) :
            oops("I need better information for %s - a number, perhaps!" % aa)
        pass

    if  (args.regex is None) and (args.not_regex is None) :
        oops("I need a regular expression!")

    if  not len(args.file_names) :
        oops("I need a log file name!")

    return(args)


def line_ended(s) :
    s   = s or ""
    if  len(s) and (s[-1] != '\n') :
        s  += "\n"                          # needed?
    return(s)


def read_log_file(log_file_name) :
    fd  = line_ended(tzlib.safe_read_whole_text_file(log_file_name + ".1"))
    fd += line_ended(tzlib.safe_read_whole_text_file(log_file_name))
    return(fd)


def make_history_file_name(log_file_name, regex, not_regex) :
    fn  = os.path.splitext(os.path.basename(log_file_name))[0]
    fn += "_" + regex
    if  not_regex :
        fn += "_not_" + not_regex
    fn  = fn.replace(".", "_")
    fn += "_%08x" % tzlib.blkcrc32(tzlib.INITIAL_CRC32_VALUE, fn)
    fn += ".known"
    fn  = tzlib.file_name_able(fn)
    fn  = os.path.join(os.path.dirname(get_config_file_name()), fn)
    return(fn)


def main(log_file_names, regex, not_regex, clear_cache = False, verbose = 0) :
    cs      = ""
    for afn in log_file_names :
        fns     = glob.glob(afn)
        for fn in fns :
            if  (os.path.splitext(afn)[1].upper() == '.GZ') or (os.path.splitext(fn)[1].upper() != '.GZ') :
                if  verbose > 1 :
                    print("Checking " + fn)
                cfd = read_log_file(fn)
                hfn = make_history_file_name(fn, regex, not_regex)
                hfd = ((not clear_cache) and tzlib.safe_read_whole_text_file(hfn)) or ""
                mfd = cfd
                if  hfd and mfd.startswith(hfd) :
                    if  verbose :
                        print("Skipping %u chars from %s from the start of %s" % ( len(hfd), hfn, fn, ))
                    mfd = cfd[len(hfd):]                # skip looking at parts of the log file we've already looked at and reported for
                cs += mfd
                tzlib.safe_makedirs(os.path.dirname(hfn))
                tzlib.safe_write_whole_text_file(   hfn, cfd)
            pass
        pass

    s   = []
    rx  = (    regex and re.compile(    regex)) or None
    nrx = (not_regex and re.compile(not_regex)) or None
    for ln in re.split(r"\r?\n", cs) :
        if  (not rx) or (rx.search(ln)) :
            if  (not nrx) or (not nrx.search(ln)) :
                s.append(ln)
            pass
        pass

    s   = "\n".join(s) + "\n"
    if  not s.strip() :
        return("")
    return(s)



if  __name__ == '__main__' :
    program_name    = sys.argv[0]
    args            = get_args(sys.argv[1:], program_name)

    s               = main(args.file_names, args.regex, args.not_regex, clear_cache = args.clear_cache, verbose = args.verbose)
    if  s           :
        print(s)
    pass

#
#
# eof
