#!/usr/bin/python

# n_point_calibration.py
#       --copyright--                   Copyright 2013 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       May 21, 2014            bar
#       March 5, 2023           bar     future print
#       --eodstamps--
##      \file
#       \namespace              tzpython.n_point_calibration
#
"""

       Calibrate an area, like a touch screen, using N points.

        From C code, with at least one bug fixed, at http://www.analog.com/static/imported-files/application_notes/AN-1021.pdf

"""

from    __future__  import  print_function

try :
    xrange
except NameError :
    xrange  = range


help_str    = """
%s (options) input_x input_y ...

    Given N calibration points, tell where an input point would be after being
    moved in a way appropriate for the calibration points.

Options:

    --point     real_x real_y input_x input_y       Where a point should be and where its input value is before calibration.

"""


class   a_point(object) :
    """ Define a point. """
    def __init__(me, x, y) :
        me.x    = x
        me.y    = y
    #   a_point


class   a_calibration_point(object) :
    """ Define a calibration point. """
    def __init__(me, real_x, real_y, input_x, input_y) :
        me.rp   = a_point( real_x,  real_y)
        me.ip   = a_point(input_x, input_y)
    #   a_calibration_point



class   a_calibrator(object) :

    """ Define a calibrator. """

    def __init__(me,  calibration_points) :
        me.points   = calibration_points
        me.KX1      = None                  # coefficients for calibration algorithm
        me.KX2      = None
        me.KX3      = None
        me.KY1      = None
        me.KY2      = None
        me.KY3      = None

    def compute_mults(me) :
        """ Compute the multipliers. """
        if  me.KX1 is None :
            a       = [ 0.0 ] * 3
            b       = [ 0.0 ] * 3
            c       = [ 0.0 ] * 3
            d       = [ 0.0 ] * 3

            if  len(me.points) == 3 :
                for pi, p  in enumerate(me.points) :
                    a[pi]   =   p.ip.x
                    b[pi]   =   p.ip.y
                    c[pi]   =   p.rp.x
                    d[pi]   =   p.rp.y
                pass

            elif len(me.points) > 3 :

                for p in me.points :
                    a[2]   += p.ip.x
                    b[2]   += p.ip.y
                    c[2]   += p.rp.x
                    d[2]   += p.rp.y
                    a[0]   += p.ip.x*p.ip.x
                    a[1]   += p.ip.x*p.ip.y
                    b[1]   += p.ip.y*p.ip.y
                    c[0]   += p.ip.x*p.rp.x
                    c[1]   += p.ip.y*p.rp.x
                    d[0]   += p.ip.x*p.rp.y
                    d[1]   += p.ip.y*p.rp.y

                b[0]        = a[1] / a[2]
                b[1]       /= b[2]
                a[0]       /= a[2]
                a[1]       /= b[2]
                c[0]       /= a[2]
                c[1]       /= b[2]
                d[0]       /= a[2]
                d[1]       /= b[2]

                a[2]       /= len(me.points)
                b[2]       /= len(me.points)
                c[2]       /= len(me.points)
                d[2]       /= len(me.points)

            k               = ((a[0] - a[2]) * (b[1] -b[2])) - ((a[1] - a[2]) * (b[0] - b[2]))

            me.KX1          = (((c[0] - c[2]) * ( b[1]         -  b[2]        ))  -  ((c[1] - c[2]) * ( b[0]         -  b[2]        ))                                             ) / k
            me.KX2          = (((c[1] - c[2]) * ( a[0]         -  a[2]        ))  -  ((c[0] - c[2]) * ( a[1]         -  a[2]        ))                                             ) / k
            me.KX3          = (( b[0]         * ((a[2] * c[1]) - (a[1] * c[2])))  +  ( b[1]         * ((a[0] * c[2]) - (a[2] * c[0])))  +  (b[2] * ((a[1] * c[0]) - (a[0] * c[1])))) / k

            me.KY1          = (((d[0] - d[2]) * ( b[1]         -  b[2]        ))  -  ((d[1] - d[2]) * ( b[0]         -  b[2]        ))                                             ) / k
            me.KY2          = (((d[1] - d[2]) * ( a[0]         -  a[2]        ))  -  ((d[0] - d[2]) * ( a[1]         -  a[2]        ))                                             ) / k
            me.KY3          = (( b[0]         * ((a[2] * d[1]) - (a[1] * d[2])))  +  ( b[1]         * ((a[0] * d[2]) - (a[2] * d[0])))  +  (b[2] * ((a[1] * d[0]) - (a[0] * d[1])))) / k

            print("@@@@", me.KX1, me.KX2, me.KX3, me.KY1, me.KY2, me.KY3)
        pass


    def find_xy(me, x, y) :
        """ Find the real location of the given input location. """
        me.compute_mults()
        rx  = (me.KX1 * x) + (me.KX2 * y) + me.KX3
        ry  = (me.KY1 * x) + (me.KY2 * y) + me.KY3
        return(rx, ry)

    def find_point(me, p) :
        me.compute_mults()
        rx  = (me.KX1 * p.x) + (me.KX2 * p.y) + me.KX3
        ry  = (me.KY1 * p.x) + (me.KY2 * p.y) + me.KY3
        return(a_point(rx, ry))

    #   a_calibrator




if  __name__ == '__main__' :

    import  os
    import  sys

    import  tzlib
    import  TZCommandLineAtFile


    program_name    = sys.argv.pop(0)

    TZCommandLineAtFile.expand_at_sign_command_line_files(sys.argv)

    points          = []
    verbose         = 0

    while True :
        oi  = tzlib.array_find(sys.argv, [ "--help", "-h", "-?", "/h", "/H", "/?" ] )
        if  oi < 0  :   break
        del sys.argv[oi]
        print(help_str % ( os.path.basename(program_name), ))
        sys.exit(254)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--verbose", "-v", ] )
        if  oi < 0  :   break
        del sys.argv[oi]
        verbose    += 1

    while True :
        oi  = tzlib.array_find(sys.argv, [ "--test9", "--t9", "-t9", ] )
        if  oi < 0  :   break
        del sys.argv[oi]
        sys.argv   += [ '-p', '3931', '3849', '3927', '3920' ]
        sys.argv   += [ '-p', '2047', '3849', '2054', '3936' ]
        sys.argv   += [ '-p',  '164', '3849',  '193', '3943' ]
        sys.argv   += [ '-p', '3931', '2047', '3911', '2119' ]
        sys.argv   += [ '-p', '2047', '2047', '2054', '2127' ]
        sys.argv   += [ '-p',  '164', '2047',  '195', '2164' ]
        sys.argv   += [ '-p', '3931',  '246', '3915',  '331' ]
        sys.argv   += [ '-p', '2047',  '246', '2050',  '354' ]
        sys.argv   += [ '-p',  '164',  '246',  '189',  '371' ]
        sys.argv   +=                       [ '3927', '3920' ]
        sys.argv   +=                       [ '2054', '3936' ]
        sys.argv   +=                       [ ' 193', '3943' ]
        sys.argv   +=                       [ '3911', '2119' ]
        sys.argv   +=                       [ '2054', '2127' ]
        sys.argv   +=                       [ ' 195', '2164' ]
        sys.argv   +=                       [ '3915',  '331' ]
        sys.argv   +=                       [ '2050',  '354' ]
        sys.argv   +=                       [ ' 189',  '371' ]


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--point", "--calibration_point", "--calibration-point", "--calibrationpoint", "-p", ] )
        if  oi < 0  :   break
        del sys.argv[oi]
        rx      = float(sys.argv.pop(oi))
        ry      = float(sys.argv.pop(oi))
        ix      = float(sys.argv.pop(oi))
        iy      = float(sys.argv.pop(oi))
        points.append(a_calibration_point(rx, ry, ix, iy))

    if  len(points) :
        me          = a_calibrator(points)
        for ai in xrange(0, len(sys.argv), 2) :
            x       = float(sys.argv[ai    ])
            y       = float(sys.argv[ai + 1])
            nx, ny  = me.find_xy(x, y)
            print("Input: %5.0f,%5.0f   Is at: %5.0f,%5.0f" % ( x, y, nx, ny, ))
        pass
    pass


#
#
# eof
