#!/usr/bin/python

# near_gps_files.py
#       --copyright--                   Copyright 2010 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       November 7, 2010
#       November 29, 2011       bar     pyflake cleanup
#       May 27, 2012            bar     doxygen namespace
#       April 9, 2016           bar     prettier printout and --wide
#       February 4, 2017        bar     --label_near
#       November 7, 2017        bar     maxint->maxsize
#       April 27, 2020          bar     handle empty GPS files
#       March 20, 2021          bar     add some cmd line option alternatives
#                                       better printout
#       July 24, 2021           bar     fix a bug in --ignore
#       July 16, 2022           bar     comments and code for cmd line altitude range, without help-mention and actually doing the job (which, as it turns out, probably isn't needed today)
#       July 30, 2022           bar     implement, sorta ok, min and max altitude filtering
#                                       use the gps file cache
#       December 7, 2022        bar     put min_altitude and max_altitude in --help
#                                       and add --max_speed
#       February 23, 2023       bar     get rid of has_keys
#       February 25, 2023       bar     python3
#       April 13, 2023          bar     listize keys values items
#       May 30, 2023            bar     allow ^C out of cache updating
#       June 5, 2023            bar     use gps_cache_file.a_file.can_point_be_inside()
#                                       take out the 1000 meter slack in min/max altitude
#       June 14, 2023           bar     able to find points near .jpg files
#       July 11, 2023           bar     MIN_SLACK
#       --eodstamps--
#       \file
#       \namespace              tzpython.near_gps_files
#
#       Tell track files with a point nearest to the given point.
#
#
#       TODO:
#           average distance of cmd line N points around the close point (like --m??_altitide #) to help filter airplane tracks)
#
#

from    __future__  import  print_function

import  difflib
import  sys

import  gps_file_cache
from    gps_file_cache  import  a_cache, a_file         # needed here for us to unpickle the cache
a_cache;a_file                                          # make pyflakes happy
import  tz_gps


MIN_SLACK   = 10.0                                      # check files that can be this close to the desired point in nautical miles


help_str    = """
python  %s   (options)   "lat,lon"  ambiguous_input_files.gpx_nmea_kml_kmz

    Tell input files with points closest to the given point.

Options:

    --max_files     n                   Report the nearest 'n' files, only                (default: 1).
    --flat                              Use flat distance rather than including altitude.
    --ignore        file_name           Do not check the given file.
    --file_since    date/time           Only input files dated since  the given date/time (default: year 1970).
    --file_before   date/time           Only input files dated before the given date/time (default: year 2038).
    --min_altitude  meters              Do not find points lower  than this.
    --max_altitude  meters              Do not find points higher than this.
    --max_speed     KPH                 Do not find points faster than this.
    --sub_dirs                          Look for files of given names in sub-directories  (and --ignore files, too).
    --wide                              Print the results widely.
    --label_near    latlon NMI  label   Know this named label for a point.
    --verbose                           Increase the verbosity (to see better files as they are found).

"""


def closest_point_in_file(point, fn, flat = False) :
    rtn = point.flat_distance_from if flat else point.distance_from

    c   = gps_file_cache.a_file(fn)
    pa  = c.add_info()                              # note: this works for our new a_file(), but would not return the points from an already-add_info'd file

    bp  = None
    bd  = sys.maxsize
    for p in pa :
        d       = rtn(p)
        if  bd  > d :
            bd  = d
            bp  = p
        pass

    return(bp)



def closest_points_by_file(point, fns, flat = False, min_altitude = -(sys.maxsize - 1), max_altitude = sys.maxsize, max_speed = sys.maxsize, fcache = None, verbose = 0) :
    rtn = point.flat_distance_from if flat else point.distance_from

    if  not isinstance(fns, (tuple, list)) :
        fns     = [ fns ]

    fnl         = max([ len(fn) for fn in fns ] + [ 15 ])
    fls         = {}
    bd          = sys.maxsize
    for fn in fns :
        if  fcache and (bd < sys.maxsize) :
            c   = fcache.get(fn)
            if  c :
                if  not c.can_point_be_inside(point.lat, point.lon, max(bd, MIN_SLACK)) :
                    continue                                                    # our point isn't in this file
                pass
            pass
        p       = closest_point_in_file(point, fn, flat)
        if  p   :
            aok = ((p.altitude is None) or (min_altitude <= p.altitude <= max_altitude)) and ((p.speed is None) or (p.speed <= max_speed))
            if  aok and verbose :
                d       = rtn(p)
                if  bd  > d :
                    bd  = d
                    print("%-*s best distance so far: %10.6f %s" % ( fnl, fn, d, p, ))
                elif verbose > 1 :
                    print("%-*s      distance       : %10.6f %s" % ( fnl, fn, d, p, ))
                pass
            if  aok     :
                fls[fn] = p
            pass
        pass

    return(fls)


def sort_files_by_closest_point(point, fns, flat = False, min_altitude = -(sys.maxsize - 1), max_altitude = sys.maxsize, max_speed = sys.maxsize, fcache = None, verbose = 0) :
    rtn = point.flat_distance_from if flat else point.distance_from

    if  not isinstance(fns, (tuple, list)) :
        fns = [ fns ]

    fls     = closest_points_by_file(point, fns, flat = flat, min_altitude = min_altitude, max_altitude = max_altitude, max_speed = max_speed, fcache = fcache, verbose = verbose)
    fns     = [ fn for fn in fns if (fn in fls) ]       # filter out files with no point close enough, so an array of file names to be in distance order, and a dict of file names to point to the closest points)

    da      = {}
    for fn in fls.keys() :
        da[fn]  = rtn(fls[fn])                          # get the distance for all the closest points in each file

    fns.sort(key = lambda a : da[a])                    # sort the file names by their closest point

    return(fns, fls, da)



def find_in_labels(labels, s) :
    """ Find the given name in the labels dictionary. """
    s       = s.lower()
    if  s  in labels :
        lb  = labels[s]
        return(lb.lat, lb.lon)

    bl      = difflib.get_close_matches(s, list(labels.keys()), 1)
    if  len(bl) :
        lb  = labels[bl[0]]
        return(lb.lat, lb.lon)

    return(None, None)


def point_can_be_in_file(fcache, fn, p, min_altitude = -(sys.maxsize - 1), max_altitude = sys.maxsize, verbose = 0) :
    """ Using a GPS file cache, return True if the given point, 'p', can be in the GPS file. """
    while True  :
        c       = fcache.get(fn)
        if  c   :
            if  ((c.min_alt is not None) and (c.min_alt > max_altitude)) or ((c.max_alt is not None) and (c.max_alt < min_altitude)) :      # note: remember, altitude is in meters
                return(False)
            return(c.can_point_be_inside(p.lat, p.lon, -0.1))
        if  verbose > 2 :
            print("Cache adding:", fn)
        fcache.add(fn)
    pass



if  __name__ == '__main__' :
    import  os

    import  TZCommandLineAtFile
    import  latlon
    import  tz_parse_time
    import  tzlib


    program_name    = sys.argv.pop(0)
    TZCommandLineAtFile.expand_at_sign_command_line_files(sys.argv)

    flat            = False
    excludes        = []
    file_before     = float(sys.maxsize)
    file_since      = float(0)
    max_altitude    = sys.maxsize
    min_altitude    = -(sys.maxsize - 1)
    max_speed       = sys.maxsize
    do_sub_dirs     = False
    max_files       = 1
    wide            = False
    verbose         = 0
    labels          = {}


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--help", "-h", "-?", "/h", "/H", "/?" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        print(help_str % os.path.basename(program_name))
        sys.exit(254)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--verbose", "--show_info", "--show-info", "--showinfo", "-v", ] )
        if  oi < 0  :    break
        del sys.argv[oi]
        verbose    += 1


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--label_near", "-l" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        s                       = sys.argv.pop(oi)
        ( lat, lon )            = latlon.parse_lat_lon(s)
        if  (lat == None) or (lon == None) :
            print("I cannot understand label_near of %s" % s)
            sys.exit(103)
        p                       = tz_gps.a_circle(lat, lon, float(sys.argv.pop(oi)))
        p.name                  = sys.argv.pop(oi).strip()
        labels[p.name.lower()]  = p


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--wide", "-w", ] )
        if  oi < 0  :    break
        del sys.argv[oi]
        wide        = True


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--flat", "-f", ] )
        if  oi < 0  :    break
        del sys.argv[oi]
        flat        = True


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--max_files", "--max-files", "--maxfiles", "-m", "-c", ] )
        if  oi < 0  :    break
        del sys.argv[oi]
        max_files   = max(int(sys.argv.pop(oi)), 1)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--ignore" ] )
        if  oi < 0  :   break
        del sys.argv[oi]
        excludes   += tzlib.ambiguous_file_list(sys.argv.pop(oi), do_sub_dirs = do_sub_dirs)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--file_since", "--file-since", "--filesince", "--file_after", "--file-after", "--fileafter", "-s", "--fs", "--fa", "--since", "--after", "-a", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        file_since              = tz_parse_time.parse_time(sys.argv.pop(oi))
        if  file_since         == None :
            print("I cannot understand the file_since!")
            sys.exit(102)
        pass


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--file_before", "--file-before", "--filebefore", "--fb", "--before", "-b", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        file_before             = tz_parse_time.parse_time(sys.argv.pop(oi))
        if  file_before        == None :
            print("I cannot understand the file_before!")
            sys.exit(102)
        pass


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--max_speed", "--max_spd", "--mxs", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        max_speed               = float(sys.argv.pop(oi))


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--max_altitude", "--max_alt", "--mxa", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        max_altitude            = float(sys.argv.pop(oi))

    while True :
        oi  = tzlib.array_find(sys.argv, [ "--min_altitude", "--min_alt", "--mna", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        min_altitude            = float(sys.argv.pop(oi))

    if  max_altitude < min_altitude :
        print("--max_altitude %f is lower than --min_altitude %f" % ( max_altitude, min_altitude, ))
        sys.exit(102)


    ifns    = {}
    for fn in excludes :
        ifns[os.path.abspath(fn)]   = True


    if  len(sys.argv) < 2 :
        print("I need a lat,lon and at least 1 file to check!")
        sys.exit(103)


    lls             = sys.argv.pop(0)
    ( lat, lon )    = latlon.parse_lat_lon(lls)
    if  lat is None :
        ( lat,  lon )   = find_in_labels(labels, lls)
        if  lat is None :
            print("I cannot understand %s!" % lls)
            sys.exit(103)
        pass
    p               = tz_gps.a_point(lat = lat, lon = lon)

    fns     = {}
    for afn in sys.argv :
        fls = tzlib.ambiguous_file_list(afn)
        for fn in fls :
            fn  = os.path.abspath(fn)
            if  fn not in ifns :
                if  file_since <= os.path.getmtime(fn) < file_before :
                    fns[fn] = True
                pass
            pass
        pass

    if  not len(fns) :
        print("No file found!", file = sys.stderr)
        sys.exit(104)


    fns         = list(fns.keys())

    fcache      = gps_file_cache.a_cache.open()
    try         :
        fns     = [ ffn for ffn in fns if point_can_be_in_file(fcache, ffn, p, min_altitude, max_altitude, verbose = verbose) ]
    except KeyboardInterrupt :
        fcache.save()
        print('')
        print("Cache saved")
        print('')
        raise
    fcache.save()

    fts         = {}
    for fn in fns :
        fts[fn] = os.path.getmtime(fn)
    fns.sort(key = lambda a : fts[a])

    ( fns, flps, flds ) = sort_files_by_closest_point(p, fns, flat = flat, min_altitude = min_altitude, max_altitude = max_altitude, max_speed = max_speed, fcache = fcache, verbose = verbose)
    if  not len(fns) :
        print("No close-enough, non-empty file found!", file = sys.stderr)
        sys.exit(104)

    fns             = fns[:max_files]
    cp              = tzlib.common_path(fns)
    cpl             = len(cp)
    if  len(fns)    < 2 :
        cpl         = 0
    bfns            = {}
    for fn in fns   :
        bfns[fn]    = fn[cpl:]              # make a file name that's the closest to the basename as it can be, making sure not to have two files by the same printed name
    mxln            = max([ len(ffn) for ffn in bfns.values() ] + [ len(' ') ])

    for fn in fns   :
        print("%-*s %15.5f" % ( mxln, bfns[fn], flds[fn], ), end = ' ')
        if  not wide :
            print('')
        print(" ", flps[fn])

    pass


#
#
# eof
