#!/usr/bin/python

# tweak_jpg_gps.py
#       --copyright--                   Copyright 2023 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       August 4, 2023          bar
#       --eodstamps--
##      \file
#       \namespace              tzpython.tweak_jpg_gps
#
#
"""
python  %s (options) input_file
Version %s

  Change the GPS point in a .jpg/.jpeg file.
    I can do multiple files, but don't ask me to.
    There would be too many things that could go wrong.

  Options:

    --lat           degrees     Add 'degrees' to the image's latitude.
    --lon           degrees     Add 'degrees' to the image's longitude.
    --alt           meters      Add 'meters'  to the image's altitude.

    --output        file_name   Output the 1 .jpg/.jpeg file to this file name.

                                Either or both of --lat and --lon
                                must be non-zero.

    --verbose                   Raise the verbosity level.

    --version                   Print the program version, current date/time
                                  and exit.

    @file_name                  Include command line options given in this file.
                                    Semi-colons start comment lines in the file.
                                Note: The file looks like this:

                                      ; ----------------------------------
                                      ; %s.py options

                                      --verbose

                                      ; eof
    Note: If the file,
    %s
    exists, I add it as an '@' file to the beginning of the command line options.

"""

from    __future__  import  print_function

import  glob
import  os
import  sys
import  time

import  tzlib
import  tz_jpg


__version__ = "0.0.0"


MY_NAME = os.path.splitext(os.path.basename(__file__))[0]
if  not MY_NAME :
    MY_NAME = "tweak_jpg_gps"


def oops(err_msg) :
    sys.stderr.write(err_msg + "\n")
    sys.exit(1)


def get_config_file_name() :
    """ Return the user's default config file name. """
    return(tzlib.expand_user_vars(os.path.join('~', '.' + MY_NAME, MY_NAME + '.cfg')))


def get_args(argv, program_name)  :
    """ Yeah, well, the Python world is so filled with getargs modules. I'd like a pop up CUI to be generated, myself, driven from the available options. And infinite history for each field and for all options as a group, and etc. """
    argv                = list(argv)

    args                = tzlib.an_object('an_args')
    args.verbose        = 0                                                                 # how printy we are
    args.args           = []                                                                # unrecognized arguments
    args.file_names     = []
    args.lat            = 0.0                                                               # how much to tweak the latitude  by
    args.lon            = 0.0                                                               # how much to tweak the longitude by
    args.alt            = 0.0                                                               # how much to tweak the altitude  by
    args.ofile_name     = None                                                              # write the output file to the original (making a .bak)

    cfg_file_name       = os.path.abspath(get_config_file_name())                           # config file to add run-time command line options to

    #
    #   Print HELP if just running the program with no options, etc.
    #       if  (not len(argv)) or (tzlib.array_find([ a.lower() for a in argv ], [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0) :
    #
    if  tzlib.find_argi_any_case(argv, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ]) >= 0 :
        bpn             = os.path.basename(program_name or (MY_NAME + ".py"))
        print(__doc__ % ( bpn, __version__, MY_NAME, cfg_file_name, ))
        sys.exit(0)                         # well, there goes my 254 :)

    import  TZCommandLineAtFile
    if  os.path.isfile(      cfg_file_name) :
        argv.insert(0, "@" + cfg_file_name)
    TZCommandLineAtFile.expand_at_sign_command_line_files(argv, encoding = 'utf8')

    while len(argv)     :
        aa  = argv.pop(0)
        a   = aa.lower()

        try :
            if  tzlib.find_argi(a, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0 :
                oops("Please keep --help /help -h -? /? -? /h out of configuration/option files!")
                sys.exit(1)

            if  tzlib.find_argi(a, [ "--verbose", "-v", ]) >= 0 :
                args.verbose       += 1
                continue

            if  tzlib.find_argi(a, [ "--version", ]) >= 0 :
                print(MY_NAME + " version " + __version__ + " run at " + time.asctime())
                sys.exit(0)

            if  tzlib.find_argi(a, [ "--lat", "--latitude",  ]) >= 0 :
                args.lat    = float(argv.pop(0))
                continue

            if  tzlib.find_argi(a, [ "--lon", "--longitude", ]) >= 0 :
                args.lon    = float(argv.pop(0))
                continue

            if  tzlib.find_argi(a, [ "--alt", "--altitude",  ]) >= 0 :
                args.alt    = float(argv.pop(0))
                continue

            if  tzlib.find_argi(a, [ "-o", "--output",  ]) >= 0 :
                args.ofile_name = argv.pop(0)
                continue

            if  a and (a[0] != '-') :
                args.file_names.append(aa)
                continue

            oops("I cannot understand %s!" % aa)

        except IndexError :
            oops("I need more information for %s !" % aa)
        except ( TypeError, ValueError, ) :
            oops("I need better information for %s - a number, perhaps!" % aa)
        pass

    if  (not args.lat) and (not args.lon) :
        oops("I need a non-zero tweak (change) in degrees to either latitude or longitude or both!")

    return(args)



if  __name__ == '__main__' :
    program_name    = sys.argv[0]
    args            = get_args(sys.argv[1:], program_name)

    fns             = {}
    for afn in args.file_names :
        fa          = glob.glob(tzlib.expand_user_vars(afn))
        for fn in fa :
            fns[os.path.abspath(fn)]    = fn
        pass

    if  args.ofile_name and (len(fns) > 1) :
        oops("I cannot write more than one file to the output file name!")

    dfns        = []
    for fn in fns.values() :
        jpg     = tz_jpg.a_jpg(fn)
        exif    = jpg.get_exif()
        if  not exif :
            oops("No exif in %s !" % fn)
        lat = exif.get_latitude()
        lon = exif.get_longitude()
        alt = exif.get_altitude()
        if  (lat is None) and (lon is None) :
            oops("No latitude or longitude in %s !" % fn)
        if  lat is not None :
            lat    += args.lat
        if  lon is not None :
            lon    += args.lon
        if  alt is not None :
            alt    += args.alt
        jpg.set_geotag(lat, lon, alt)

        tzone           = 0
        mtime           = None
        if  True        :                                               # tz_jpg.py's --use_gps_time option
            try         :
                mtime   = jpg.get_gps_time()
            except      :
                mtime   = None

            if  (mtime or 0) > 0 :
                jpg.set_time(mtime)
                print("Note: No GPS time in ", fn, file = sys.stderr)
            jpg.save(args.ofile_name or fn)
            args.ofile_name = None
            if  jpg.set_file_date_time_from_exif(tzone) <= 0 :      # !!!! does not use gps time if an xml time is given
                print("Could not set date/time for", fn, file = sys.stderr)
            else    :
                dfns.append(fn)
            pass
        pass

    if  not len(dfns) :
        oops("I cannot find any .jpg files to do!")

    pass

#
#
# eof
