#!/usr/bin/python
# -*- coding: utf8 -*-

# tz_androsensor.py
#       --copyright--                   Copyright 2023 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       March 30, 2023          bar
#       April 13, 2023          bar     listize keys values items
#       June 6, 2023            bar     use tz_gps.a_track()
#                                       only print exception when verbose
#       July 11, 2023           bar     don't be so eager to print column-not-found
#       October 8, 2023         bar     figure out androsensor doesn't do nmea lat/lon and we should use the lat/lon column data
#                                       read .csv files from the zip files I put them in after a day or two
#                                       ll_dupe_filter for ...../gps/x.py to sense time when the device is flat or unmoving or whatever
#                                       bring in the ACCELEROMETER values because they work on roam and move, both
#                                       V0.0.2
#                                       kinda fix the local times to GMT/UTC (as best we can - we'll think the date-times are now on the local machine. Ug.
#       --eodstamps--
##      \file
#       \namespace              tzpython.tz_androsensor
#
#
#       WARNING!    gps/r.py has a comment indicating I've found making GPS tracks from AndroSensor files is a problem. Time stamp problems and big jumps here and there. Or something. See the comment.
#
#
"""
python  %s (options) input_file(s)...
Version %s

  Read mobile phone app AndroSensor .csv files.

  Options:

    --verbose                   Raise the verbosity level.

    --version                   Print the program version, current date/time
                                  and exit.

    @file_name                  Include command line options given in this file.
                                    Semi-colons start comment lines in the file.
                                Note: The file looks like this:

                                      ; ----------------------------------
                                      ; %s.py options

                                      --verbose

                                      ; eof
    Note: If the file,
    %s
    exists, I add it as an '@' file to the beginning of the command line options.

"""

from    __future__  import  print_function

import  difflib
import  os
import  re
import  sys
import  time
import  zipfile

import  tzlib
import  tz_gps
import  tz_parse_time


__version__ = "0.0.2"


MY_NAME = os.path.splitext(os.path.basename(__file__))[0]
if  not MY_NAME :
    MY_NAME = "tz_androsensor"


def oops(err_msg) :
    sys.stderr.write(err_msg + "\n")
    sys.exit(1)


def get_config_file_name() :
    """ Return the user's default config file name. """
    return(tzlib.expand_user_vars(os.path.join('~', '.' + MY_NAME, MY_NAME + '.cfg')))


def get_args(argv, program_name)  :
    """ Yeah, well, the Python world is so filled with getargs modules. I'd like a pop up CUI to be generated, myself, driven from the available options. And infinite history for each field and for all options as a group, and etc. """
    argv                = list(argv)

    args                = tzlib.an_object('an_args')
    args.verbose        = 0                                                                 # how printy we are
    args.args           = []                                                                # unrecognized arguments
    args.file_names     = []

    cfg_file_name       = os.path.abspath(get_config_file_name())                           # config file to add run-time command line options to

    #
    #   Print HELP if just running the program with no options, etc.
    #       if  (not len(argv)) or (tzlib.array_find([ a.lower() for a in argv ], [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0) :
    #
    if  tzlib.find_argi_any_case(argv, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ]) >= 0 :
        bpn             = os.path.basename(program_name or (MY_NAME + ".py"))
        print(__doc__ % ( bpn, __version__, MY_NAME, cfg_file_name, ))
        sys.exit(0)                         # well, there goes my 254 :)

    import  TZCommandLineAtFile
    if  os.path.isfile(      cfg_file_name) :
        argv.insert(0, "@" + cfg_file_name)
    TZCommandLineAtFile.expand_at_sign_command_line_files(argv, encoding = 'utf8')

    while len(argv)     :
        aa  = argv.pop(0)
        a   = aa.lower()

        try :
            if  tzlib.find_argi(a, [ "--help", "/help", "-h", "-?", "/?", "-?", "/h", ] ) >= 0 :
                oops("Please keep --help /help -h -? /? -? /h out of configuration/option files!")
                sys.exit(1)

            if  tzlib.find_argi(a, [ "--verbose", "-v", ]) >= 0 :
                args.verbose       += 1
                continue

            if  tzlib.find_argi(a, [ "--version", ]) >= 0 :
                print(MY_NAME + " version " + __version__ + " run at " + time.asctime())
                sys.exit(0)

            if  a and (a[0] != '-') :
                args.file_names.append(aa)
                continue

            oops("I cannot understand %s!" % aa)

        except IndexError :
            oops("I need more information for %s !" % aa)
        except ( TypeError, ValueError, ) :
            oops("I need better information for %s - a number, perhaps!" % aa)
        pass

    return(args)


androsensor_hdrs        = {
                            'ACCELEROMETER X (m/s²)'                           :   'accel_x',
                            'ACCELEROMETER Y (m/s²)'                           :   'accel_y',
                            'ACCELEROMETER Z (m/s²)'                           :   'accel_z',
                            'GRAVITY X (m/s²)'                                 :   'gravity_x',
                            'GRAVITY Y (m/s²)'                                 :   'gravity_y',
                            'GRAVITY Z (m/s²)'                                 :   'gravity_z',
                            'LINEAR ACCELERATION X (m/s²)'                     :   'lin_accel_x',
                            'LINEAR ACCELERATION Y (m/s²)'                     :   'lin_accel_y',
                            'LINEAR ACCELERATION Z (m/s²)'                     :   'lin_accel_z',
                            'GYROSCOPE X (rad/s)'                               :   'gyro_rad_x',
                            'GYROSCOPE Y (rad/s)'                               :   'gyro_rad_y',
                            'GYROSCOPE Z (rad/s)'                               :   'gyro_rad_z',
                            'LIGHT (lux)'                                       :   'lux',
                            'MAGNETIC FIELD X (μT)'                            :   'mag_x',
                            'MAGNETIC FIELD Y (μT)'                            :   'mag_y',
                            'MAGNETIC FIELD Z (μT)'                            :   'mag_z',
                            'ORIENTATION Z (azimuth °)'                        :   'orient_azimuth_z',
                            'ORIENTATION X (pitch °)'                          :   'orient_pitch_x',
                            'ORIENTATION Y (roll °)'                           :   'orient_roll_y',
                            'PROXIMITY (i)'                                     :   'proximity',
                            'ATMOSPHERIC PRESSURE (hPa)'                        :   'pressure_hPa',
                            'SOUND LEVEL (dB)'                                  :   'sound_db',
                            'LOCATION Latitude :'                               :   'lat',
                            'LOCATION Longitude :'                              :   'lon',
                            'LOCATION Altitude ( m)'                            :   'alt',
                            'LOCATION Altitude-google ( m)'                     :   'alt_google',
                            'LOCATION Altitude-atmospheric pressure ( m)'       :   'alt_pressure',
                            'LOCATION Speed ( Mph)'                             :   'mph',
                            'LOCATION Accuracy ( m)'                            :   'location_accuracy_m',
                            'LOCATION ORIENTATION (°)'                         :   'location_orientation',
                            'Satellites in range'                               :   'satellite_counts',
                            'GPS NMEA'                                          :   'nmea',
                            'Time since start in ms'                            :   'mills',
                            'YYYY-MO-DD HH-MI-SS_SSS'                           :   'time',                     # local time on move and roam October 9, 2023
                          }
for s in list(androsensor_hdrs.keys()) :                            # note: list() must be here because we're deleting items inside the loop
    androsensor_hdrs[tzlib.convert_to_unicode(s).encode('utf8')]    = androsensor_hdrs[s]


def get_androsensor_column_names(hdr, verbose = 0) :
    """ Given the 1st line of the .csv file, return an array of column names corresponding to the comma delimited "columns". """
    if  not hdr :
        return([])

    hdrs    = hdr.split(',')
    if  len(hdrs) < 2 :
        return([])

    cols    = []
    for ci, s in enumerate(hdrs) :
        ss  = tzlib.convert_to_unicode(s).encode('utf8')
        ms  = difflib.get_close_matches(ss, list(androsensor_hdrs.keys()), 1)
        if  len(ms) != 1 :
            if  verbose > 2 :
                print("Cannot find AndroSensor column %s!" % tzlib.convert_to_unicode(s))
            return([])

        cols.append(androsensor_hdrs[ms[0]])

    return(cols)


def num_or_none(s) :
    if  not s :
        return(None)
    return(float(s.strip()))


def time_or_none(s) :
    if  not s :
        return(None)
    return(tz_parse_time.parse_time(s.strip()))


def process_nmea_strs(d) :
    """ Break out the NMEA string information given in the AndroSensor .cvs format. Add the info to the given dictionary. """
    ns      = d.get('nmea', None)
    if  ns  :
        na  = [ '$' + s for s in ns.lstrip('$').split('$') ]
        for gs in na :
            g       = re.search(r'^\$([a-zA-Z]+),', gs)
            if  g   :
                d['nmea_' + g.group(1).lower()] = gs
                gpgsa           = tz_gps.a_gpgsa.from_string(gs)
                if  gpgsa       :
                    d['gpgsa']  = gpgsa
                pass
            pass
        pass
    pass


class   an_androsensor_point(tz_gps.a_point) :

    def __init__(me, ln, col_names, tzone = None) :
        oln         = ln
        flds        = []
        while len(ln) :
            ci      = ln.find('"')
            ei      = ln[ci + 1:].find('"')
            if  (ci < 0) or (ln[max(0, ci - 1)] != ',') or (ln[min(len(ln) - 1, ci + 1 + ei + 1)] != ',') :     # we don't do double-quotes at the start of the current string (two successive double-quote fields, for instance) or ending double quotes at the end of the string
                if  ln.endswith(',') :
                    ln  = ln[:-1]
                flds   += ln.split(',')
                break
            if  ei  < 0 :
                raise ValueError('No closing double-quote %s' % oln)
            ei     += ci + 1
            flds   += ln[:ci - 1].split(',')
            flds.append(ln[ci + 1 : ei])
            ln      = ln[ei + 2:]                                           # skip over the comma after the ending double-quote

        if  len(flds) != len(col_names) :
            if  (len(flds) == 31) and (len(col_names) == 32) :
                col_names   = list(col_names)
                for i, cn in enumerate(col_names) :
                    if  cn == 'sound_db' :
                        del(col_names[i])
                        break
                    pass
                pass
            pass
        if  len(flds) != len(col_names) :
            raise ValueError('Wrong number of columns, not %u, is %u in %s %s' % ( len(col_names), len(flds), oln, repr(flds), ))

        d           = dict(zip(col_names, flds))

        when        = time_or_none(d.get('time',            None))

        speed       = num_or_none(d.get('mph',              None))
        if  speed  != None :
            speed  /= tzlib.KILOMETERS_TO_MILES                             # convert mph to kph

        lat         = num_or_none(d.get('lat',              None))          # cell tower lat/lon - nmea data has the real values
        lon         = num_or_none(d.get('lon',              None))
        alt         = num_or_none(d.get('alt',              None))

        me.tzone    = tzone

        me.nmea_point   = None
        process_nmea_strs(d)
        pnts            = tz_gps.points_from_all_tracks(tz_gps.tracks_from_nmea_file_or_string('\n'.join([ d[ns] for ns in d.keys() if ns.startswith('nmea_') ]) + '\n'))
        if  len(pnts)  == 1 :                                                                           # the nmea fields don't have the same tod and lat/lon values - they are not consistent, so we sometimes don't get a nmea point
            me.nmea_point  = pnts[0]                                                                    # since this point has nmea string, use that info because it's more authoritative
            lat         = me.nmea_point.lat
            lon         = me.nmea_point.lon
            alt         = me.nmea_point.altitude
            me.tzone    = int(((me.nmea_point.when - when) + (7.5 * 60)) / (15 * 60)) * 15 * 60         # nmea is in zulu time, the normal stuff is local. We want zulu, but the local time has the mills. So we'll add the time zone to the local time.
        elif not me.tzone :
            me.tzone    = int(when - tz_parse_time.parse_time(time.asctime(time.localtime(when))))      # figure out the offset to GMT from local time on this computer, now. The .csv should be in GMT, but isn't.

        if  me.tzone   != None :
            when       += me.tzone

        super(an_androsensor_point, me).__init__(when       = when,
                                                 lat        = lat or 0.0,
                                                 lon        = lon or 0.0,
                                                 altitude   = alt or 0.0,
                                                 speed      = speed or 0.0,
                                                 sat_cnt    = num_or_none(d.get('satellite_counts', '').split('/')[0]),
                                                )
        me.androsensor_fields   = d

        me.proximity            = num_or_none(me.androsensor_fields.get('proximity',        None))
        me.sound_db             = num_or_none(me.androsensor_fields.get('sound_db',         None))
        me.azimuth              = num_or_none(me.androsensor_fields.get('orient_azimuth_z', None))
        me.pitch                = num_or_none(me.androsensor_fields.get('orient_pitch_x',   None))
        me.roll                 = num_or_none(me.androsensor_fields.get('orient_roll_y',    None))      # 180 when face down, 0 when face up
        me.accel_x              = num_or_none(me.androsensor_fields.get('accel_x',    None))            # for sensing non-movement. October 9, 2023 both phones show ACCELEROMETER.
        me.accel_y              = num_or_none(me.androsensor_fields.get('accel_y',    None))
        me.accel_z              = num_or_none(me.androsensor_fields.get('accel_z',    None))            # -9.8 when face down, 9.8 when face up (depending on camera bump, etc.)
        # . !!!!
        # .
        # .

        me.compass              = me.azimuth            # 0..359.XXXX - true north or magnetic?

        # print("@@@@ %.6f" % when, d.get('time',            None), me.when)

        # print("@@@@", me.nmea_point, me.lat, me.lon, me.altitude, me.speed)
        # !!!! put the sensor data in places of our own - somehow get them to GPX files
        pass


    #   an_androsensor_point


def csv_lines_have_nmea(lns, cols, verbose = 0) :
    """ Do these .csv data lines have NMEA data in them anywhere? """
    for ln in lns   :
        try         :
            if  len(ln.strip()) :
                p   = an_androsensor_point(ln, cols)
                if  p.nmea_point != None :

                    return(True)    # ok, the file has NMEA data

                pass
            pass
        except ValueError :         # for instance, too many values to unpack in FILE_NAME
            if  verbose > 1 :
                tzlib.print_exception()
            pass
        pass
    return(False)


def extract_androsensor_points(csv_text, file_name = None, ll_dupe_filter = True, verbose = 0) :
    """ Given an androsensor file in text form, get an array of an_androsensor_point()s. """
    pts             = []
    lns             = re.split('\r?\n', csv_text)
    if  len(lns)    < 3 :
        if  verbose > 1 :
            print("Nothing in %s!" % file_name)
        pass
    else            :
        cols        = get_androsensor_column_names(lns[0])
        tzone       = None
        dlns        = lns[1:]
        have_nmea   = csv_lines_have_nmea(dlns, cols, verbose = verbose)
        pp                  = None
        for ln in dlns      :
            try             :
                if  len(ln.strip()) :
                    p       = an_androsensor_point(ln, cols, tzone = tzone)
                    tzone   = p.tzone
                    if  (not have_nmea) or (not ll_dupe_filter) or (p.nmea_point != None) :                         # if we have nmea and this point does not, then ignore this point. If there's no $GPGGA, then there's no real info - the location info from the tower should not be convoluted with GPS location
                        if  (not pp)    or (not ll_dupe_filter) or (pp.lat != p.lat) or (pp.lon != p.lon) or (pp.altitude != p.altitude) :      # AndroSensor can record the same point for multiple seconds - ignore such things - GPX Logger simply doesn't record points in the same situation
                            pts.append(p)                                                                           # if the file doesn't have nmea data, we strip lat/lon/alt dupes. October 8, 2023 sigh.
                        pp  = p                                                                                     # note: it's not certain whether this should be inside or outside the None check
                    pass
                pass
            except ValueError as e :
                if  verbose :
                    print(e, "in", file_name)
                pass
            pass
        pass

    return(pts)


def read_csvs_from_zip(fn) :
    """ Read all .csv files from a zip file, if possible. Returns array of [ file_name, csv_text ] in file-name-sorted order. """
    retval  = []
    if  os.path.splitext(fn)[1].upper() == '.ZIP' :
        zf  = zipfile.ZipFile(fn, "r", zipfile.ZIP_DEFLATED)
        fns = [ fnm for fnm in zf.namelist() if os.path.splitext(fnm)[1].upper() == '.CSV' ]
        fns.sort(key  = lambda s : (s.upper(), s, ))
        for fn in fns :
            fd  = tzlib.convert_to_unicode(zf.read(fn))
            # print("@@@@", fn, len(re.split(r'\r?\n', fd)))
            retval.append([ fn, fd, ])
        zf.close()
    return(retval)


def read_file(fn, ll_dupe_filter = True, verbose = 0) :
    """ Return tracks in tz_gps array form, created from an AndroSensor .csv file. The raw values are in the point.androsensor_fields dictionary. """
    file_list       = read_csvs_from_zip(fn) or [ [ fn, tzlib.safe_read_whole_text_file(fn), ], ]
    tracks          = []                        # we'll combine the tracks together from all the .gpx files
    for fn, fd in file_list :
        points      = extract_androsensor_points(fd, file_name = fn, ll_dupe_filter = ll_dupe_filter, verbose = verbose)
        if  points  :
            # print("@@@@", fn, len(re.split(r'\r?\n', fd)), "lines", len(points), "pts", points[0], "to", points[-1])
            tracks.append(tz_gps.a_track(points, name = os.path.basename(fn), file_name = fn))
        pass
    tracks  = tz_gps.make_array_of_tracks(tracks)
    return(tracks)



if  __name__ == '__main__' :
    program_name    = sys.argv[0]
    args            = get_args(sys.argv[1:], program_name)

    tracks          = []
    for fn in sorted(args.file_names) :
        tracks     += read_file(fn, verbose = args.verbose)

    print("%u tracks read with %u points" % ( len(tracks), len(tz_gps.points_from_all_tracks(tracks))))

#
#
# eof
