#!/usr/bin/python

# tz_xtrinsic.py
#       --copyright--                   Copyright 2015 (C) Tranzoa, Co. All rights reserved.    Warranty: You're free and on your own here. This code is not necessarily up-to-date or of public quality.
#       --url--                         http://www.tranzoa.net/tzpython/
#       --email--                       pycode is the name to send to. tranzoa.com is the place to send to.
#       --bodstamps--
#       May 14, 2015            bar
#       May 15, 2015            bar     what a mess. just do it simple some time, keeping in mind the apparent way the board sends CRLF at the beginning of output text lines rather than the end - and its unfriendliness to type-ahead or even fast typing
#       November 7, 2017        bar     maxint->maxsize
#       March 5, 2023           bar     future print
#       --eodstamps--
##      \file
#       \namespace              tzpython.tz_xtrinsic
#
#
"""
    Communicate with a Freescale XTRINSIC dev board.
"""


from    __future__  import  print_function

import  os
import  re
import  sys
import  time

import  serial              # from pySerial

import  tz_usb
import  tzlib


USB_VENDOR_ID   = 0x1357
USB_PRODUCT_ID  = 0x0707



SAMP_TRY_INTERVAL   = 9


class   a_comm_exception(Exception) :
        pass

class   a_comm_timeout_exception(a_comm_exception) :
        pass                                # there's nothing to be had for now




class   a_comm :

    def __init__(me, io = None, timeout = 0.01, verbose = 0) :
        me.io           = io
        me.timeout      = timeout
        me.ibuf         = ""
        me.scnt         = 1
        me.data         = []                    # queue of a_recording's driven by data coming in from the device
        me.busy         = tzlib.elapsed_time()
        me.verbose      = verbose or 0

        if  not me.io   :
            me.reopen()
        pass


    def rx(me, how_many = 1) :
        try :
            return(me.io.read(how_many))
        except ( serial.SerialException, serial.serialutil.SerialException, OSError, IOError, AttributeError ) :
            pass
        return('')


    def read(me, how_many = 1, timeout = None) :
        timeout         = timeout or me.timeout

        t               = nt    = tzlib.elapsed_time()
        r               = ""
        while True :
            rr          = me.rx(how_many)
            if  len(rr) :
                me.busy = tzlib.elapsed_time()
            how_many   -= len(rr)
            rc          = len(rr)
            r          += rr
            if  not how_many :
                break

            if  rc      :
                t       = nt

            if  nt - t >= timeout :
                if  len(r) :
                    break
                raise a_comm_timeout_exception("timeout")

            nt          = tzlib.elapsed_time()

            if  not rc  :
                time.sleep(min(0.1, timeout / 2.0))
            pass

        # print("rxing", len(r), "%04x" % ( len(r) ), hexify(r))

        return(r)


    def readline(me) :
        while True  :
            try     :
                r   = me.read()
            except a_comm_timeout_exception :
                break
            if  not len(r) :
                break
            me.ibuf += r
        r           = ""
        i           = me.ibuf.find('\n')

        # print("@@@@[%s]" % me.ibuf, i)

        if  i >= 0  :
            r       = me.ibuf[:i]
            r       = r.strip()
            me.ibuf = me.ibuf[i + 1:]

        return(r)


    def write(me, s) :
        if  s   :
            me.busy = tzlib.elapsed_time()
            if  me.verbose > 2 :
                print("Wrt", repr(s))
            try     :
                me.io.write(s)
                me.io.flush()
                me.ibuf = ""
                while me.rx() :
                    pass
                pass
            except ( OSError, IOError, AttributeError, serial.SerialException ) :
                raise a_comm_exception("write")
            pass
        pass


    def cmd(me, cmd) :
        me.write("\r")
        me.start_usb()
        me.write(cmd)
        time.sleep(.2)
        me.write("\r")


    def is_ready(me) :
        return(tzlib.elapsed_time() - me.busy >= 2.5)


    def close(me)   :
        if  me.io   :
            c       = me.io
            me.io   = None
            try     :
                c.close()
            except  :
                raise a_comm_exception("close")
            pass
        pass



    def blind_close(me) :
        try :
            me.close()
        except a_comm_exception :
            pass
        pass


    def reopen(me, port = None) :
        me.busy         = tzlib.elapsed_time()
        me.blind_close()

        if  not port    :
            port        = tz_usb.find_likely_COM_ports(vendor_id = USB_VENDOR_ID, product_id = USB_PRODUCT_ID)
            if  port    :
                port    = port[0]
            pass

        if  not port :
            return(None)

        try :
            port    = int(port)
            cport   = port - 1
        except ValueError :
            cport   = port

        try :
            me.io   = serial.Serial(port = cport, baudrate = 115200, timeout = 0.001)                # note: PC program sets 8O1. serial.Serial() multiplies timeout by 1000 before passing to windows (This 1 mill is minimum for windows. I don't know about other OS's.)
        except serial.SerialException :
            return(False)

        return(True)


    def start_usb(me) :
        while True  :
            t       = tzlib.elapsed_time() - 5
            try     :
                while not me.is_ready() :
                    tt  = tzlib.elapsed_time()
                    if  tt - t >= 5 :
                        t       = tt
                        me.write("\r")
                    time.sleep(.2)
                    # print("starting", repr(me.readline()))
                break
            except ( OSError, IOError, AttributeError, serial.SerialException, a_comm_exception ) :
                print("@@@@@@@@@@@@@@@@@@@@@@@@@@@")
                me.reopen()
            pass
        pass

    pass        # a_comm




def hms(t) :
    it  = int(t)
    return("%2u:%02u:%04.1f" % ( it / 3600, (it / 60) % 60, t - ((it / 60) * 60) ) )




def main(ident) :
    import      TZCommandLineAtFile
    import      TZKeyReady

    program_name    = sys.argv.pop(0)

    TZCommandLineAtFile.expand_at_sign_command_line_files(sys.argv)


    port        = 0             # my COM port, not yours
    port_list   = 0
    interval    = 0
    verbose     = 0


    help_str    = """
%s (options) (output_file)

    I record data from the Freescale XTRINSIC evaluation board.

Options:

    --port  port_number     Set the COM port number
    --port_list             List possible ports (twice, list all available ports)
    --interval  seconds     Print all 3 measurement types and append them to the file.
    --version               Print the program version number.


""" % ( os.path.basename(program_name) )


    oi  = tzlib.array_find(sys.argv, [ "--help", "-?", "?", "-h", "/h", "/?", "?" ] )
    if  oi >= 0 :
        print(help_str)
        sys.exit(254)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--port", "-p" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        if  (oi >= len(sys.argv)) or not len(sys.argv[oi]) :
            print("No COM port given!")
            sys.exit(102)
        port        = sys.argv.pop(oi)


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--port_list" ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        port_list  += 1

    while True :
        oi  = tzlib.array_find(sys.argv, [ "--interval", "--poll", "-t", "-i", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        interval    = float(sys.argv.pop(oi))


    while True :
        oi  = tzlib.array_find(sys.argv, [ "--verbose", "-v", ] )
        if  oi < 0 :    break
        del sys.argv[oi]
        verbose    += 1



    ofile_name      = None
    if  len(sys.argv) >= 1 :
        ofile_name  = sys.argv.pop(0)

        if  ofile_name.startswith('-') :
            print("Put the whole path or a dot/slash before the output file name. Dashes are confusing: [%s]" % ofile_name)
            sys.exit(104)

        ofile_name  = tzlib.expand_user_vars(ofile_name)


    if  len(sys.argv) :
        print("I don't understand %s (--help for options)!" % ( sys.argv ))
        sys.exit(104)



    if  port_list :
        print("Likely COM ports, if any:")
        print(tz_usb.find_likely_COM_ports(vendor_id = USB_VENDOR_ID, product_id = USB_PRODUCT_ID, list_level = port_list))
        print("")


    if  not port :
        port    = tz_usb.find_likely_COM_ports(vendor_id = USB_VENDOR_ID, product_id = USB_PRODUCT_ID)
        if  port :
            print(  "Using port", port[0],)
            if  len(port) > 1 :
                print("Found ports", port, end = ' ')
            print('')
            port    = port[0]
        pass

    if  not port :
        print("Please tell me a COM port to use with the --port option (e.g. --port 2 or --port /dev/ttyACM0 )!")
        sys.exit(103)

    try :
        port    = int(port)
        cport   = port - 1
    except ValueError :
        cport   = port

    try :
        io  = serial.Serial(port = cport, baudrate = 115200, timeout = 0.001)       # note: PC program sets 8O1. serial.Serial() multiplies timeout by 1000 before passing to windows (This 1 mill is minimum for windows. I don't know about other OS's.)
        io.close()
    except  :   # serial.SerialException :
        print("Port %s [%s] cannot be opened!" % ( str(port), str(cport) ))
        sys.exit(111)

    interval    = interval or sys.maxsize

    me          = a_comm(verbose = verbose)

    me.start_usb()              # in case he's not turned it on (though we'll do this every half second of silence from the device, anyway

    print("Type ? for help")

    cmds        = {
                    "MPL3115"   :   "s0",
                    "MAG3110"   :   "s1",
                    "MMA8491Q"  :   "s2",
                  }
    rcmds       = {}

    t           = tzlib.elapsed_time()
    when        = t - interval
    do_samp     = t - SAMP_TRY_INTERVAL
    try         :
        while True  :
            try     :
                li  = me.readline()
                t   = tzlib.elapsed_time()
                if  t - when >= interval :
                    do_samp = t - SAMP_TRY_INTERVAL
                    when   += interval
                if  do_samp :
                    ping    = False
                    typ     = re.sub(r':.*$', '', li)
                    if  (typ in cmds) and (typ not in rcmds) :
                        do_samp     = t
                        rcmds[typ]  = li
                        if  len(rcmds) < len(cmds) :
                            ping    = True
                        else        :
                            tm      = time.time()
                            if  ofile_name :
                                fo  = open(ofile_name, 'at')
                            for s in rcmds.values() :
                                tl  = "%s %17.6f %s" % ( time.asctime(time.localtime(tm)), tm, s, )
                                if  ofile_name :
                                    fo.write(tl + "\n")
                                print(tl)
                            if  ofile_name :
                                fo.close()
                            rcmds   = {}
                            do_samp = 0
                        pass
                    elif t - do_samp >= SAMP_TRY_INTERVAL :
                        ping        = True

                    if  ping        :
                        for typ in cmds.keys() :
                            if  typ not in rcmds :
                                me.cmd(cmds[typ])
                                break
                            pass
                        do_samp     = tzlib.elapsed_time()

                    if  li          :
                        # print("@@@@ sample done", li, typ)
                        pass
                    pass
                pass
            except a_comm_exception :
                # tzlib.print_exception()
                # print("@@@@ reopening for comm error")
                me.reopen()
                # print("@@@@ reopend   for comm error")
                time.sleep(3.0)
                rcmds       = {}
                t           = tzlib.elapsed_time()
                when        = t - interval
                do_samp     = t - SAMP_TRY_INTERVAL

            if  not li      :
                time.sleep(0.2)

            while True :
                k   = TZKeyReady.tz_key_ready()
                if  not k : break
                if  k == '\n' : k   = '\r'
                me.write(k)
            pass
        pass
    except KeyboardInterrupt :
        print('')

    me.blind_close()


if  __name__ == '__main__' :

    main("")


#
#
# eof
